# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""Utilities related to calling subprocesses
"""
from subprocess import run, PIPE, CalledProcessError
import os
from os.path import abspath, expanduser, join, isfile
import click

from dataworkspaces.errors import ConfigurationError


def call_subprocess(args, cwd, verbose=False):
    """Call an executable as a child process. Returns the standard output.
    If it fails, we will print
    an error and allow CalledProcessError to be thrown.
    """
    if verbose:
        click.echo(" ".join(args) + " [run in %s]" % cwd)
    cp = run(args, cwd=cwd, encoding="utf-8", stdout=PIPE, stderr=PIPE)
    try:
        cp.check_returncode()
    except CalledProcessError:
        click.echo(cp.stderr)
        raise
    if verbose:
        click.echo(cp.stdout)
    return cp.stdout  # can ignore if you don't need this.


def call_subprocess_for_rc(args, cwd, verbose=False):
    """Call an executable as a child process. Returns the return
    code of the call.
    """
    if verbose:
        click.echo(" ".join(args) + " [run in %s]" % cwd)
    cp = run(args, cwd=cwd, encoding="utf-8", stdout=PIPE, stderr=PIPE)
    if verbose:
        click.echo(cp.stdout)
        click.echo("%s exited with %d" % (args[0], cp.returncode))
    return cp.returncode


STANDARD_EXE_SEARCH_LOCATIONS = ["/usr/bin", "/usr/local/bin", abspath(expanduser("~/bin"))]


def find_exe(
    exe_name, recommended_action_on_error, additional_search_locations=STANDARD_EXE_SEARCH_LOCATIONS
):
    """Find an executable, first checking in the current path
    (as would be done by the shell), and then searching including
    any directories specified by additional_search_locations.
    If found, returns the full path to the executable. Otherwise,
    raises a ConfigurationError
    """
    dirpaths = os.environ["PATH"].split(os.pathsep)
    for l in additional_search_locations:
        if l not in dirpaths:
            dirpaths.append(l)
    for dirpath in dirpaths:
        fpath = join(dirpath, exe_name)
        if isfile(fpath) and os.access(fpath, os.X_OK):
            return fpath
    raise ConfigurationError(
        "Did not find executable '%s'. Tried searching in: %s. %s"
        % (exe_name, ", ".join(dirpaths), recommended_action_on_error)
    )



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
import os
from os.path import abspath, expanduser, join
from typing import List, Optional

import click

from dataworkspaces.workspace import init_workspace, RESOURCE_ROLE_CHOICES
from dataworkspaces.errors import ConfigurationError


def init_command(
    name: str,
    hostname: str,
    create_resources: List[str],
    scratch_dir: Optional[str] = None,
    git_fat_remote: Optional[str] = None,
    git_fat_user: Optional[str] = None,
    git_fat_port: Optional[int] = None,
    git_fat_attributes: Optional[str] = None,
    git_lfs_attributes: Optional[str] = None,
    batch: bool = False,
    verbose: bool = False,
):
    if git_fat_attributes and git_lfs_attributes:
        raise ConfigurationError("Cannot specify git-fat and git-lfs for the same repository")
    workspace_dir = abspath(expanduser(os.curdir))
    if scratch_dir is None:
        scratch_dir = join(workspace_dir, "scratch")
    workspace = init_workspace(
        "dataworkspaces.backends.git",  # TODO: remove hardcoding
        name,
        hostname,
        batch,
        verbose,
        scratch_dir,
        workspace_dir,
        git_fat_remote,
        git_fat_user,
        git_fat_port,
        git_fat_attributes,
        git_lfs_attributes,
    )

    if len(create_resources) > 0:
        click.echo("Will now create sub-directory resources for " + ", ".join(create_resources))
        for role in create_resources:
            assert role in RESOURCE_ROLE_CHOICES, "bad role name %s" % role
            workspace.add_resource(
                role,
                "git-subdirectory",
                role,
                join(workspace_dir, role),
                confirm_subdir_create=False,
            )
        click.echo("Finished initializing resources:")
        for role in create_resources:
            click.echo("  %s: ./%s" % (role, role))

    workspace.save("workspace initialization")
    click.echo("Workspace %s initialized successfully." % name)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
File-related utilities
"""

import os
from os.path import dirname, isdir, abspath, expanduser, exists, isabs, commonpath, isfile, join
import shutil
import click
from typing import Optional

from dataworkspaces.errors import ConfigurationError


def remove_dir_if_empty(path: str, base_dir: str, verbose: bool = False) -> None:
    """Remove an empty directory and any parents that are empty,
    up to base_dir.
    """
    if path == base_dir:
        return
    elif len(os.listdir(path)) == 0:
        os.rmdir(path)
        if verbose:
            print("Removing (now) empty directory %s" % path)
        remove_dir_if_empty(dirname(path), base_dir, verbose)


def safe_rename(src: str, dest: str) -> None:
    """Safe replacement for os.rename(). The problem is that os.rename()
    does not work across file systems. In that case, you need to actually
    copy the file
    """
    try:
        os.rename(src, dest)
    except OSError:
        try:
            if isdir(src):
                shutil.copytree(src, dest)
                shutil.rmtree(src)
            else:
                shutil.copyfile(src, dest)
                os.remove(src)
        except Exception as e:
            raise ConfigurationError("Unable to copy %s to %s: %s" % (src, dest, e)) from e


def get_subpath_from_absolute(absolute_parent_path: str, absolute_child_path: str) -> Optional[str]:
    """Given two absolute paths where one is the parent of the other, return the
    child path as a relative path from the parent. Returns None if the paths are
    equal and raises a ValueError if absolute_child_path is not actually a
    parent of absolute_parent_path.
    """
    assert isabs(absolute_parent_path)
    assert isabs(absolute_child_path)
    if absolute_parent_path.endswith("/") and len(absolute_parent_path) > 1:
        absolute_parent_path = absolute_parent_path[:-1]
    if absolute_child_path.endswith("/") and len(absolute_child_path) > 1:
        absolute_child_path = absolute_child_path[:-1]
    if absolute_child_path == absolute_parent_path:
        return None
    if commonpath([absolute_parent_path, absolute_child_path]) != absolute_parent_path:
        raise ValueError(
            "'%s is not a subpath of '%s'" % (absolute_child_path, absolute_parent_path)
        )
    else:
        return absolute_child_path[len(absolute_parent_path) + 1 :]


def does_subpath_exist(
    base_dir: str, subpath: str, must_be_file: bool = False, must_be_directory: bool = False
) -> bool:
    path = join(base_dir, subpath)
    if isfile(path) and (not must_be_directory):
        return True  # includes links
    elif isdir(path) and not (must_be_file):
        return True
    else:
        return False  # does not exist, but a special file


def parent_path(path):
    """Return the path to the parent directory of path"""
    return abspath(join(path, os.pardir))


class LocalPathType(click.Path):
    """A subclass of click's Path input parameter type used to validate a local path
    where we are going to put a resource. The path does not necessarily exist yet, but
    we need to validate that the parent directory exists and is writable.

    If must_be_outside_of_workspace is set to the workspace directory, we validate
    that the path is outside of the workspace.

    If allow_multiple_levels_of_missing_dirs is True, we go up the tree to find the
    first parent in the hierarchy that exists and check its permissions. This works
    in cases where we're going to use os.makedirs() to create multiple levels of
    the hierarchy.
    """

    def __init__(
        self,
        exists: bool = False,
        must_be_outside_of_workspace: Optional[str] = None,
        allow_multiple_levels_of_missing_dirs: bool = False,
    ):
        super().__init__(
            exists=exists,
            file_okay=False,
            dir_okay=True,
            writable=True,
            # always use unicode, not bytes
            path_type=str,
        )
        self.must_be_outside_of_workspace = must_be_outside_of_workspace
        self.allow_multiple_levels_of_missing_dirs = allow_multiple_levels_of_missing_dirs

    def convert(self, value, param, ctx):
        rv = abspath(expanduser(super().convert(value, param, ctx)))
        if isdir(rv):
            parent_dir = rv
        else:
            parent_dir = parent_path(rv)
        if self.allow_multiple_levels_of_missing_dirs:
            while not isdir(parent_dir) and parent_dir != "/":
                parent_dir = parent_path(parent_dir)
        if not exists(parent_dir):
            self.fail('%s "%s" does not exist.' % (self.path_type, parent_dir), param, ctx)  # type: ignore
        if not isdir(parent_dir):
            self.fail('%s "%s" is a file.' % (self.path_type, parent_dir), param, ctx)  # type: ignore
        if not os.access(parent_dir, os.W_OK):
            self.fail('%s "%s" is not writable.' % (self.path_type, parent_dir), param, ctx)  # type: ignore
        if (self.must_be_outside_of_workspace is not None) and commonpath(
            [self.must_be_outside_of_workspace, rv]
        ) in (self.must_be_outside_of_workspace, rv):
            self.fail(
                '%s must be outside of workspace "%s"'
                % (self.path_type, self.must_be_outside_of_workspace),  # type: ignore
                param,
                ctx,
            )
        return rv



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
from typing import Optional, List, cast
import click

from dataworkspaces.errors import ConfigurationError, UserAbort, InternalError, ApiParamError
from dataworkspaces.workspace import (
    Workspace,
    SnapshotWorkspaceMixin,
    SnapshotResourceMixin,
    ResourceRoles,
)


def restore_command(
    workspace: Workspace,
    tag_or_hash: str,
    only: Optional[List[str]] = None,
    leave: Optional[List[str]] = None,
    strict: bool = False,
) -> int:
    """Run the restore and return the number of resources affected.
    """
    if not isinstance(workspace, SnapshotWorkspaceMixin):
        raise ConfigurationError("Workspace %s does not support snapshots" % workspace.name)
    mixin = cast(SnapshotWorkspaceMixin, workspace)
    # First, find the history entry
    md = mixin.get_snapshot_by_tag_or_hash(tag_or_hash)

    # process the lists of resources
    current_names = set(workspace.get_resource_names())
    # get the non-null resources in snapshot
    snapshot_names = set(
        [rn for rn in md.restore_hashes.keys() if md.restore_hashes[rn] is not None]
    )
    all_names = current_names.union(snapshot_names)
    if (only is not None) and (leave is not None):
        raise ApiParamError("Cannot specify both only and leave for restore command.")
    elif only is not None:
        # For only, we will be a little stricter, as the user is explicitly
        # specifying the resources.
        restore_set = set(only)
        strict = True
    elif leave is not None:
        restore_set = all_names.difference(leave)
    else:
        restore_set = all_names

    # We need to remove result resources from the restore set, as we
    # do not restore them to their prior state.
    result_resources = {
        rname
        for rname in restore_set
        if workspace.get_resource_role(rname) == ResourceRoles.RESULTS
    }
    result_resources_in_restore_set = result_resources.intersection(restore_set)
    if len(result_resources_in_restore_set) > 0:
        if strict:
            raise ConfigurationError(
                "Restore set contains result resources, which cannot be restored. The following are result resources: %s"
                % ", ".join(result_resources_in_restore_set)
            )
        else:
            click.echo(
                "Skipping the restore of the following result resources, which are left in their latest state: %s"
                % ", ".join(result_resources_in_restore_set)
            )
            restore_set = restore_set.difference(result_resources)

    # error checking
    invalid = restore_set.difference(all_names)
    if len(invalid) > 0:
        raise ConfigurationError("Resource name(s) not found: %s" % ", ".join(sorted(invalid)))
    removed_names = restore_set.difference(current_names)
    if len(removed_names) > 0:
        if strict:
            raise ConfigurationError(
                "Resources have been removed from workspace or have no restore hash and strict mode is enabled."
                + " Removed resources: %s" % ", ".join(sorted(removed_names))
            )
        else:
            click.echo(
                "Skipping restore of resources that have been removed from workspace or have no restore hash: %s"
                % ", ".join(sorted(removed_names)),
                err=True,
            )
            restore_set = restore_set.difference(removed_names)
    added_names = restore_set.difference(snapshot_names)
    if len(added_names) > 0:
        if strict:
            raise ConfigurationError(
                "Resources have been added to workspace since restore, and strict mode enabled."
                + " Added resources: %s" % ", ".join(sorted(added_names))
            )
        else:
            click.echo(
                "Resources have been added to workspace since restore, will leave them as-is: %s"
                % ", ".join(sorted(added_names)),
                err=True,
            )
            restore_set = restore_set.difference(added_names)

    # get ordered list of names and resources as well as restore hashes
    restore_name_list = [rn for rn in workspace.get_resource_names() if rn in restore_set]
    if len(restore_name_list) == 0:
        click.echo("No resources to restore.")
        return 0
    restore_resource_list = [workspace.get_resource(rn) for rn in restore_name_list]
    for r in restore_resource_list:
        if not isinstance(r, SnapshotResourceMixin):
            raise InternalError(
                "Resource %s was in snapshot, but is not a SnapshotResourceMixin" % r.name
            )
    restore_hashes = {rn: md.restore_hashes[rn] for rn in restore_set}

    tagstr = " (%s)" % ",".join(md.tags) if len(md.tags) > 0 else ""
    click.echo("Restoring snapshot %s%s" % (md.hashval, tagstr))

    def fmt_rlist(rnames):
        if len(rnames) > 0:
            return ", ".join(rnames)
        else:
            return "None"

    click.echo("  Resources to restore: %s" % fmt_rlist(restore_name_list))
    names_to_leave = sorted(current_names.difference(restore_set))
    click.echo("  Resources to leave: %s" % fmt_rlist(names_to_leave))
    if not workspace.batch:
        # Unless in batch mode, we always want to ask for confirmation
        resp = input("Should I perform this restore? [Y/n]")
        if resp.lower() != "y" and resp != "":
            raise UserAbort()

    # do the work!
    mixin.restore(
        md.hashval, restore_hashes, cast(List[SnapshotResourceMixin], restore_resource_list)
    )
    workspace.save("Restore to %s" % md.hashval)

    return len(restore_name_list)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.

import click
from dataworkspaces.workspace import clone_workspace


def clone_command(
    backend: str, hostname: str, batch: bool = False, verbose: bool = False, *args
) -> None:
    workspace = clone_workspace(backend, hostname, batch, verbose, *args)
    click.echo(
        "Completed initial clone of workspace %s, will check for resources to clone..."
        % workspace.name
    )

    rnames = [name for name in workspace.get_names_of_resources_with_local_state()]
    if len(rnames) == 0:
        click.echo("No resources with local state to clone.")
    else:
        click.echo("Will clone the following resources: %s" % ", ".join(rnames))
        for rname in rnames:
            workspace.clone_resource(rname)

    workspace.save("Clone")
    click.echo("Successfully completed clone of workspace %s." % workspace.name)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.


import click

from dataworkspaces.errors import ConfigurationError
from dataworkspaces.workspace import Workspace


def add_command(scheme: str, role: str, name: str, workspace: Workspace, *args):
    current_names = set(workspace.get_resource_names())
    if workspace.batch:
        if name == None:
            name = workspace.suggest_resource_name(scheme, role, *args)
        else:
            if name in current_names:
                raise ConfigurationError("Resource name '%s' already in use" % name)
    else:
        suggested_name = None
        while (name is None) or (name in current_names):
            if suggested_name == None:
                suggested_name = workspace.suggest_resource_name(scheme, role, *args)
            name = click.prompt(
                "Please enter a short, unique name for this resource", default=suggested_name
            )
            if name in current_names:
                click.echo("Resource name '%s' already in use." % name, err=True)

    workspace.add_resource(name, scheme, role, *args)
    workspace.save("add of %s" % name)
    click.echo("Successful added resource '%s' to workspace." % name)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Welcome to Data Workspaces. To get started, see the documentation at
https://data-workspaces-core.readthedocs.io/en/latest/. For most applications,
you will interact through the command line interface (dws), the integrations
with common libraries like Scikit-learn (the modules dataworkspaces.kits.*), or the
lineage API (the module dataworkspaces.lineage).

Some other submodules (mostly for extending dataworkspaces):

* backends  - backends for storing workspace metadata (currently only git)
* commands  - implementations of the command line commands
* dws       - command line interface and option parsing
* errors    - exception classes
* resources - implementations of the different resource types
* utils     - common utilities
* workspace - core APIs for dataworkspaces internals, including Workspace,
              Resource, and the various mixins that represent different
              capabilities.

Submodules are not imported by default, due to external dependencies
(e.g. click or specific frameworks required by submodules under
dataworkspaces.kits). Just import the APIs you need directly. For
example::

  from dataworkspaces import lineage

or::

  from dataworkspaces.lineage import LineageBuilder

"""

__version__ = "1.5.2"




# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Definition of configuration parameters
"""
import socket
from typing import Dict, Callable, Any, Optional, Tuple, List, cast

assert Dict
assert Callable
from os.path import isabs, expanduser, abspath, join
import click
import json

from dataworkspaces.utils.snapshot_utils import validate_template
from dataworkspaces.utils.file_utils import LocalPathType, get_subpath_from_absolute
from dataworkspaces.errors import ConfigurationError, InternalError
from dataworkspaces.utils.regexp_utils import HOSTNAME_RE


class ParamNotFoundError(ConfigurationError):
    pass


class ParamValidationError(ConfigurationError):
    pass


class ParamParseError(ConfigurationError):
    pass


class UknownParamError(ConfigurationError):
    pass


class ParamType:
    """Base class for a defining the type of the parameter.
    Types are designed to be reusable across multiple parameters."""

    def parse(self, str_value: str) -> Any:
        """Convert the string representation into the
        actual value. If the parameter's type is a string, parsing must
        be itempotent - if parse is called multiple times, the same
        parsed value is returned.

        This function is used when parsing from the command line and potentally
        when parsing from JSON.
        """
        return str_value

    def validate(self, value: Any) -> None:
        """Check the value and throw a ParamValidationError if value
        is not of this type. Does not need to handle optional values,
        this is handled by the ParamDef class."""
        pass

    def to_json(self, value: Any) -> Any:
        """Convert the parameter value to a JSON-serializable value.
        Usually, you can just use the default, which returns the value as-is.
        In some cases, like dates, you might need a conversion function.
        """
        return value

    def __repr__(self):
        return self.__class__.__name__ + "()"

    def __str__(self):
        """Override this to return a more useful type name."""
        return self.__class__.__name__


class BoolType(ParamType):
    def parse(self, str_value: str) -> bool:
        if str_value.lower() == "true":
            return True
        elif str_value.lower() == "false":
            return False
        else:
            raise ParamParseError(
                "Unable to parse boolean parameter value was '%s'" % repr(str_value)
            )

    def validate(self, value: Any) -> None:
        if not isinstance(value, bool):
            raise ParamValidationError("Parameter must be a bool, value was '%s'" % repr(value))

    def __str__(self):
        return "bool"


class StringType(ParamType):
    def validate(self, value: Any) -> None:
        if not isinstance(value, str):
            raise ParamValidationError("Value '%s' is not a string" % repr(value))

    def __str__(self):
        return "string"


class AbspathType(ParamType):
    def validate(self, value: Any) -> None:
        if not isinstance(value, str):
            raise ParamValidationError(
                "Path value '%s' should be a string, not %s" % (repr(value), type(value))
            )
        if not isabs(value):
            raise ParamValidationError("Path value '%s' is not absolute" % repr(value))

    def __str__(self):
        return "absolute_path"


class RelpathType(ParamType):
    def validate(self, value: Any) -> None:
        if not isinstance(value, str):
            raise ParamValidationError(
                "Path value '%s' should be a string, not %s" % (repr(value), type(value))
            )
        if isabs(value):
            raise ParamValidationError(
                "Path value '%s' is absolute, should be relative" % repr(value)
            )

    def __str__(self):
        return "relative_path"


class HostnameType(ParamType):
    def validate(self, name):
        if not HOSTNAME_RE.match(name):
            raise ParamValidationError(
                "'%s' is not a valid hostname: must start with a letter and only contain letters, numbers, '-', '_', and '.'"
                % name
            )

    def __str__(self):
        return "hostname"


class EnumType(ParamType):
    """String parameter that has one of a fixed set of values."""

    def __init__(self, *values):
        self.values = [v.lower() for v in values]  # type: List[str]

    def parse(self, str_value: str) -> Any:
        """Convert the string representation into the
        actual value. In this case, we just ensure that it is
        normalized to lowercase.
        """
        return str_value.lower()

    def validate(self, value: Any) -> None:
        if not isinstance(value, str):
            raise ParamValidationError(
                "Enum value '%s' should be a string, not %s" % (repr(value), type(value))
            )
        if value.lower() not in self.values:
            raise ParamValidationError(
                "Enum value '%s' invalid. Must be one of %s" % (value, ", ".join(self.values))
            )

    def __repr__(self):
        return "EnumTaram(%s)" % ", ".join(self.values)

    def __str__(self):
        return self.__repr__()


class ParamDef:
    def __init__(
        self,
        name: str,
        default_value: Any,
        optional: bool,
        help: str,
        ptype: ParamType,
        allow_missing: bool = False,
    ):
        """Define a parameter - used for both workspace config params
        and resource params.

        If optional is True, that means that the parameter can take
        a value of None. If optional is False and the default_value is None,
        then the parameter must be explicitly specified. Also, set optional to
        False if you never want to allow a value of None (user cannot override
        default value to None).

        Set allow_missing to True if this parameter cannot be None but may be missing
        in a workspace created using an older version of DWS. Will just apply the
        default value and print a warning.
        """
        self.name = name
        self.default_value = default_value
        self.optional = optional
        self.help = help
        self.ptype = ptype
        if default_value is not None:
            # we validate our default value if it is specified
            ptype.validate(default_value)
        self.allow_missing = allow_missing
        if allow_missing and (default_value is None):
            raise InternalError("Parameter %s: Allow missing requires a default value" % name)

    def parse(self, raw_value: Any) -> Any:
        """If the value is a string, parse it and return the parsed value. Otherwise,
        just return the original value.

        NOTE: if the parameter's type is a string, and a parse function is defined,
        the parse function must return the value if a parsed value is passed in again.
        """
        if isinstance(raw_value, str):
            try:
                return self.ptype.parse(raw_value)
            except Exception as e:
                raise ParamParseError(
                    "Unable to parse parameter %s value %s" % (self.name, repr(raw_value))
                ) from e
        else:
            return raw_value

    def validate(self, value: Any) -> None:
        """Validate required values and then check the types's validation method"""
        if value is None and self.optional:
            return  # if None and None is allowed, we can just skip validation
        elif value is None and self.default_value is None and not self.optional:
            raise ParamValidationError("Required parameter '%s' is None" % self.name)
        elif value is None and not self.optional and self.allow_missing:
            click.echo(
                "WARNING: Parameter %s not set. Using default value of '%s'"
                % (self.name, repr(self.default_value)),
                err=True,
            )
        else:
            try:
                self.ptype.validate(value)
            except Exception as e:
                raise ParamValidationError(
                    "Error in validation of configuration parameter '%s'" % self.name
                ) from e

    def to_json(self, value: Any) -> Any:
        """Convert the parameter value to a JSON-serializable value, via the type's function.
        """
        return self.ptype.to_json(value)

    def __repr__(self):
        return "ParamDef(%s, %s, %s, %s, %s, %s)" % (
            self.name,
            repr(self.default_value),
            self.optional,
            repr(self.help),
            repr(self.ptype),
            self.allow_missing,
        )


PARAM_DEFS = {}  # type: Dict[str, ParamDef]
LOCAL_PARAM_DEFS = {}  # type: Dict[str, ParamDef]

#########################################################
#                  Global Params                        #
#########################################################
# These params are stored in the workspace's config.json
# file, which is keep in git and synced across all instances
# of the workspace.


def define_param(name: str, default_value: Any, optional: bool, help: str, ptype: ParamType) -> str:
    global PARAM_DEFS
    assert name not in PARAM_DEFS
    assert name not in LOCAL_PARAM_DEFS  # don't want duplicates across local and global
    PARAM_DEFS[name] = ParamDef(name, default_value, optional, help, ptype)
    return name


class TemplateType(ParamType):
    def validate(self, value: Any) -> None:
        validate_template(value)

    def __str__(self):
        return "template"


RESULTS_DIR_TEMPLATE = define_param(
    "results.dir_template",
    default_value="snapshots/{HOSTNAME}-{TAG}",
    optional=False,
    help="Template describing where results files will be moved during snapshot",
    ptype=TemplateType(),
)


class FileListType(ParamType):
    def parse(self, str_value: str) -> Any:
        return json.loads(str_value)

    def validate(self, value: Any) -> None:
        if not isinstance(value, list) and not isinstance(value, tuple):
            raise ParamValidationError("File list parameter must be a list or a tuple")
        for v in value:
            if not isinstance(v, str):
                raise ParamValidationError("File list parameter elements must be strings")

    def __str__(self):
        return "file_list"


RESULTS_MOVE_EXCLUDE_FILES = define_param(
    "results.move_exclude_files",
    default_value=["README.txt", "README.rst", "README.md"],
    optional=False,
    help="List of files to exclude when moving results to a subdirectory during snapshot.",
    ptype=FileListType(),
)


SCRATCH_DIRECTORY = define_param(
    "scratch_directory",
    default_value=None,
    optional=True,
    help="Directory where scratch files are stored (checkpoints, temporary data, etc.). "
    + "If this is set, it is relative to the workspace directory.",
    ptype=RelpathType(),
)


DWS_GIT_BRANCH = define_param(
    "dws_git_branch",
    default_value=None,
    optional=True,
    help="Name of git branch for the main dws repository. This is set when the dws workspace is initialized to the "+
         "current branch.",
    ptype=StringType()
)

def get_global_param_defaults():
    """Return a mapping of all default values of global params for use
    in generating the initial config file
    """
    return {n: d.default_value for (n, d) in PARAM_DEFS.items() if d.default_value is not None}


#########################################################
#                  Local Params                         #
#########################################################
# These are parameters local to the current install, and not
# tracked through the workspace.


def define_local_param(
    name: str, default_value: Optional[Any], optional: bool, help: str, ptype: ParamType
):
    global LOCAL_PARAM_DEFS
    assert name not in LOCAL_PARAM_DEFS
    assert name not in PARAM_DEFS  # don't want duplicates across local and global
    LOCAL_PARAM_DEFS[name] = ParamDef(name, default_value, optional, help, ptype)
    return name


def get_local_param_defaults(hostname=None):
    """Return a mapping of all default values for use in generating
    the initial config file. The hostname is usually provided explicitly
    by the user
    """
    defaults = {
        param: defn.default_value
        for (param, defn) in LOCAL_PARAM_DEFS.items()
        if defn.default_value is not None
    }
    if hostname is not None:
        defaults[HOSTNAME] = hostname
    return defaults


DEFAULT_HOSTNAME = socket.gethostname().split(".")[0]
HOSTNAME = define_local_param(
    "hostname",
    default_value=DEFAULT_HOSTNAME,
    optional=False,
    help="Hostname to identify this machine in snapshots.",
    ptype=HostnameType(),
)

LOCAL_SCRATCH_DIRECTORY = define_local_param(
    "local_scratch_directory",
    default_value=None,
    optional=True,
    help="Directory where scratch files are stored (checkpoints, temporary data, etc.). "
    + "If this is set, it is absolute and only specific to the local machine.",
    ptype=AbspathType(),
)


def init_scratch_directory(
    scratch_dir: str,
    workspace_dir: str,
    global_params: Dict[str, Any],
    local_params: Dict[str, Any],
) -> Tuple[str, Optional[str]]:
    """Given the user-provided or default scratch directory, set the SCRATCH_DIRECTORY
    and LOCAL_SCRATCH_DIRECTORY parameters accordingly. One only will be set, with preference
    to the global parameter, which is relative to the workspace. Returns a tuple of the absolute
    and the gitignore entry (if any) for the scratch_directory
    """
    abs_scratch_dir = abspath(expanduser(scratch_dir)) if not isabs(scratch_dir) else scratch_dir
    scratch_dir_gitignore = None  # type: Optional[str]
    if abs_scratch_dir.startswith(workspace_dir):
        rel_scratch_dir = get_subpath_from_absolute(workspace_dir, abs_scratch_dir)
        global_params[SCRATCH_DIRECTORY] = rel_scratch_dir  # always store a relative directory
        # scratch dir gitignore should start with / to indicate that this should only
        # match the exact path relative to the git repo root.
        if rel_scratch_dir is None:
            raise ConfigurationError(
                "Scratch directory cannot be equal to workspace directory. "
                + "It should either be a subdirectory or completely outside it."
            )
        if rel_scratch_dir.startswith("./"):
            scratch_dir_gitignore = rel_scratch_dir[1:]
        else:
            scratch_dir_gitignore = "/" + rel_scratch_dir
    else:
        local_params[LOCAL_SCRATCH_DIRECTORY] = abs_scratch_dir
    return (abs_scratch_dir, scratch_dir_gitignore)


def clone_scratch_directory(
    workspace_dir: str,
    global_params: Dict[str, Any],
    local_params: Dict[str, Any],
    batch: bool = False,
) -> str:
    """Set the scratch directory parameters for a cloned copy of the workspace,
    updating local_params if neded.
    Returns the absolute path for the scratch directory on this system.
    """
    if SCRATCH_DIRECTORY in global_params:
        return join(workspace_dir, global_params[SCRATCH_DIRECTORY])
    elif not batch:
        local_path = cast(
            str,
            click.prompt(
                "Please specify a location for this workspace's scratch directory (must be outside of workspace)",
                type=LocalPathType(exists=False, must_be_outside_of_workspace=workspace_dir),
            ),
        )
        local_params[LOCAL_SCRATCH_DIRECTORY] = local_path
        return local_path
    else:
        # TODO: come up with a standard way of handling this when called from the API - either by
        # letting the user provide values in advance or by having some standard defaults.
        raise ConfigurationError(
            "Scratch directory was not within workspaces and we are running in batch mode. No way to ask user for location."
        )


def get_scratch_directory(
    workspace_dir: str, global_params: Dict[str, Any], local_params: Dict[str, Any]
) -> Optional[str]:
    """Given the global and local params, return the absolute path
    of the scratch directory for this workspace. If it was not specified
    in either, print a warning and return None.
    """
    if SCRATCH_DIRECTORY in global_params:
        # normalize the path to remove any "." in the path
        return abspath(join(workspace_dir, global_params[SCRATCH_DIRECTORY]))
    elif LOCAL_SCRATCH_DIRECTORY in local_params:
        return local_params[LOCAL_SCRATCH_DIRECTORY]
    else:
        click.echo(
            "WARNING: Neither the %s nor %s parameters are set, so cannot find scratch directory. Please set one using 'dws config'."
            % (SCRATCH_DIRECTORY, LOCAL_SCRATCH_DIRECTORY),
            err=True,
        )
        return None


#########################################################
#                 Resource Params                       #
#########################################################
# These are utility functions for defining and managing the
# parameters of individual resources.


class ResourceParams:
    def __init__(self):
        self.global_defs = {}  # type: Dict[str, ParamDef]
        self.local_defs = {}  # type: Dict[str, ParamDef]
        # define the globals present in all resources
        self.define(
            "resource_type",
            default_value=None,
            optional=False,
            is_global=True,
            help="Type of this resource (e.g. git, local_files, api, etc.)",
            ptype=StringType(),
        )
        self.define(
            "name",
            default_value=None,
            optional=False,
            is_global=True,
            help="Name of the resource",
            ptype=StringType(),
        )
        self.define(
            "role",
            default_value=None,
            optional=False,
            is_global=True,
            help="Resource role (source-data, intermediate-data, code, results)",
            ptype=StringType(),
        )

    def define(
        self,
        name: str,
        default_value: Any,
        optional: bool,
        help: str,
        is_global: bool,
        ptype: ParamType,
        allow_missing: bool = False,
    ):
        assert name not in self.global_defs
        assert name not in self.local_defs
        defn = ParamDef(name, default_value, optional, help, ptype, allow_missing=allow_missing)
        if is_global:
            self.global_defs[name] = defn
        else:
            self.local_defs[name] = defn

    def get(self, name: str, supplied_value: Any) -> Any:
        """Use this in the resource's __init__ method to get
        the parameter for setting its value as an attribute. For example:

            self.export = self.get('export', params.get('export')) # type: bool
        """
        if name in self.global_defs:
            defn = self.global_defs[name]
        elif name in self.local_defs:
            defn = self.local_defs[name]
        else:
            raise ParamNotFoundError(name)
        parsed_value = defn.parse(supplied_value)
        defn.validate(parsed_value)
        return parsed_value

    def get_params(self, resource) -> Dict[str, Any]:
        """Retrieve the enclosing resource's global parameters
        and convert to json form.
        """
        return {
            name: self.global_defs[name].to_json(getattr(resource, name))
            for name in self.global_defs.keys()
        }

    def get_local_params(self, resource) -> Dict[str, Any]:
        """Retrieve the enclosing resource's local parameters
        and convert to json form.
        """
        return {
            name: self.local_defs[name].to_json(getattr(resource, name))
            for name in self.local_defs.keys()
        }



"""Utilities for git repos that use git-fat for support of large
files.
"""
from os.path import join, exists, isdir
from typing import Optional, cast

from dataworkspaces.errors import ConfigurationError
from dataworkspaces.utils.git_utils import is_git_repo, git_add, git_commit
from dataworkspaces.utils.subprocess_utils import find_exe
from dataworkspaces.utils.regexp_utils import RSYNC_RE, FPATH_RE, USERNAME_RE


def get_dot_gitfat_file_path(workspace_dir: str) -> str:
    return join(workspace_dir, ".gitfat")


def is_a_git_fat_repo(repo_dir: str) -> bool:
    assert is_git_repo(repo_dir), "%s is not a git repo" % repo_dir
    return exists(get_dot_gitfat_file_path(repo_dir))


def has_git_fat_been_initialized(repo_dir: str) -> bool:
    return isdir(join(repo_dir, ".git/fat"))


# Utility funtions for issue #12 - if a repo is git-fat enabled, and git-fat is not in the path,
# git add will fail silently for filter calls (e.g. in git add). We explicitly check that
# the executable is in the path in situations where we will call git as a subprocess.

GIT_FAT_ERRMSG = "Ensure that the dataworkspaces package is installed and that you have activated your virtual environment (if any)."


def validate_git_fat_in_path() -> None:
    """Validate that git-fat is in the path, asssuming we already know that this
    is a git-fat repo.
    If the executable is not found, throw a configuration error. We need to do this, as git itself
    will not return an error return code if a filter (e.g. git-fat) is not found.
    """
    find_exe("git-fat", GIT_FAT_ERRMSG, additional_search_locations=[])


def validate_git_fat_in_path_if_needed(repo_dir: str) -> None:
    """Validate that git-fat is in the path, if this repo is git-fat enabled.
    Otherwise, throw a configuration error. We need to do this, as git itself
    will not return an error return code if a filter (e.g. git-fat) is not found.
    """
    if not is_a_git_fat_repo(repo_dir):
        return
    find_exe("git-fat", GIT_FAT_ERRMSG, additional_search_locations=[])


def run_git_fat_pull_if_needed(repo_dir: str, verbose: bool) -> None:
    """When restoring resources from snapshots or running a pull, we need to
    also run git-fat pull, if the repo is a git-fat repo
    """
    if not is_a_git_fat_repo(repo_dir):
        return
    else:
        import dataworkspaces.third_party.git_fat as git_fat

        python2_exe = git_fat.find_python2_exe()
        git_fat.run_git_fat(python2_exe, ["pull"], cwd=repo_dir, verbose=verbose)


def run_git_fat_push_if_needed(repo_dir: str, verbose: bool) -> None:
    """Push the git-fat updates for the repo, if it is git fat-enabled
    """
    if not is_a_git_fat_repo(repo_dir):
        return
    else:
        import dataworkspaces.third_party.git_fat as git_fat

        python2_exe = git_fat.find_python2_exe()
        git_fat.run_git_fat(python2_exe, ["push"], cwd=repo_dir, verbose=verbose)


def setup_git_fat_for_repo(
    repo_dir: str,
    git_fat_remote: str,
    git_fat_user: Optional[str] = None,
    git_fat_port: Optional[int] = None,
    git_fat_attributes: Optional[str] = None,
    verbose: bool = False,
) -> None:
    """Setup git fat and all the associated configuration files
    for a repository
    """
    validate_git_fat_in_path()
    dot_git_fat_fpath = get_dot_gitfat_file_path(repo_dir)
    files_to_add = [
        ".gitfat",
    ]
    dot_git_attributes_fpath = None  # type: Optional[str]
    if git_fat_attributes:
        dot_git_attributes_fpath = join(repo_dir, ".gitattributes")
        files_to_add.append(".gitattributes")
    if (RSYNC_RE.match(git_fat_remote) is None) and (FPATH_RE.match(git_fat_remote) is None):
        raise ConfigurationError(
            (
                "'%s' is not a valid remote address for rsync (used by git-fat). "
                + "Please use the format HOSTNAME:/PATH"
            )
            % git_fat_remote
        )
    if git_fat_user is not None and USERNAME_RE.match(git_fat_user) is None:
        raise ConfigurationError("'%s' is not a valid remote username for git-fat" % git_fat_user)
    import dataworkspaces.third_party.git_fat as git_fat

    python2_exe = git_fat.find_python2_exe()
    # click.echo("Initializing git-fat with remote %s" % git_fat_remote)
    with open(dot_git_fat_fpath, "w") as f:
        f.write("[rsync]\nremote = %s\n" % git_fat_remote)
        if git_fat_user:
            f.write("sshuser = %s\n" % git_fat_user)
        if git_fat_port:
            f.write("sshport = %s\n" % git_fat_port)
    if git_fat_attributes is not None:
        with open(cast(str, dot_git_attributes_fpath), "w") as f:
            for extn in git_fat_attributes.split(","):
                f.write("%s filter=fat -crlf\n" % extn)
    git_fat.run_git_fat(python2_exe, ["init"], cwd=repo_dir, verbose=verbose)
    git_add(repo_dir, files_to_add, verbose)
    git_commit(repo_dir, "initialized git-fat with remote %s" % git_fat_remote, verbose)



"""
A Python wrapper for rclone.
"""
# Original code:
# https://github.com/ddragosd/python-rclone
#
# Modifications:
# Rupak Majumdar 2019.01.01
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=W0102,W0703,C0103

import logging
import subprocess
import tempfile

from dataworkspaces.errors import ConfigurationError

class RCloneException(ConfigurationError):
    pass

class RClone:
    """
    Wrapper class for rclone.
    """

    def __init__(self, cfgfile=None, cfgstring=None):
        self.log = logging.getLogger("RClone")
        self._ensure_rclone_exists()
        self.cfgstring = ''
        self.cfgfile = None
        if cfgstring:
            self.cfgstring = cfgstring.replace("\\n", "\n")
        elif cfgfile:
            self.cfgfile = cfgstring.replace("\\n", "\n")
        else:
            # find the default config file used by the rclone installation
            ret = self._execute(['rclone', 'config', 'file'])
            self.log.debug(ret)
            if ret['code'] == 0:
                # rclone config file output looks like:
                #
                # Configuration file is stored at:
                # filename
                # so we skip until the '\n'
                self.cfgfile = ret['out'].splitlines()[1].decode('utf_8')
            else:
                print(ret)
                raise ConfigurationError("RClone requires either a configuration file or a configuration string")

        assert(self.cfgstring or self.cfgfile), 'Either a config string is given or a filename is given'

    def _ensure_rclone_exists(self):
        ret = self._execute(['rclone', 'version'])
        if ret['code'] == -20:
            raise ConfigurationError("rclone executable not found")

    def _execute(self, command_with_args):
        """
        Execute the given `command_with_args` using Popen
        Args:
            - command_with_args (list) : An array with the command to execute,
                                         and its arguments. Each argument is given
                                         as a new element in the list.
        """
        self.log.debug("Invoking : %s", " ".join(command_with_args))
        try:
            with subprocess.Popen(
                    command_with_args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE) as proc:
                (out, err) = proc.communicate()

                #out = proc.stdout.read()
                #err = proc.stderr.read()

                self.log.debug(out)
                if err:
                    self.log.warning(err.decode("utf-8").replace("\\n", "\n"))

                return {
                    "code": proc.returncode,
                    "out": out,
                    "error": err
                }
        except FileNotFoundError as not_found_e:
            self.log.error("Executable not found. %s", not_found_e)
            return {
                "code": -20,
                "error": not_found_e
            }
        except Exception as generic_e:
            self.log.exception("Error running command. Reason: %s", generic_e)
            return {
                "code": -30,
                "error": generic_e
            }

    def run_cmd(self, command, extra_args=[]):
        """
        Execute rclone command
        Args:
            - command (string): the rclone command to execute.
            - extra_args (list): extra arguments to be passed to the rclone command
        """
        if self.cfgfile: # pass the file name to rclone
            command_with_args = ["rclone", command, "--config", self.cfgfile]
            command_with_args += extra_args
            command_result = self._execute(command_with_args)
            return command_result
     
        # config is given in a string 
        # save the configuration in a temporary file
        # and invoke rclone with the temporary file
        with tempfile.NamedTemporaryFile(mode='wt', delete=True) as cfg_file:
            # cfg_file is automatically cleaned up by python
            self.log.debug("rclone config: ~%s~", self.cfgstring)
            cfg_file.write(self.cfgstring)
            cfg_file.flush()

            command_with_args = ["rclone", command, "--config", cfg_file.name]
            command_with_args += extra_args
            command_result = self._execute(command_with_args)
            cfg_file.close()
            return command_result

    def copy(self, source, dest, flags=[]):
        """
        Executes: rclone copy source:path dest:path [flags]
        Args:
        - source (string): A string "source:path"
        - dest (string): A string "dest:path"
        - flags (list): Extra flags as per `rclone copy --help` flags.
        """
        return self.run_cmd(command="copy", extra_args=[source] + [dest] + flags)

    def sync(self, source, dest, flags=[]):
        """
        Executes: rclone sync source:path dest:path [flags]
        Args:
        - source (string): A string "source:path"
        - dest (string): A string "dest:path"
        - flags (list): Extra flags as per `rclone sync --help` flags.
        """
        return self.run_cmd(command="sync", extra_args=[source] + [dest] + flags)

    def listremotes(self, flags=[]):
        """
        Executes: rclone listremotes [flags]
        Args:
        - flags (list): Extra flags as per `rclone listremotes --help` flags.
        """
        ret = self.run_cmd(command="listremotes", extra_args=flags)
        if ret['code'] == 0:
            list_remotes = map((lambda s: s[0:-1].decode('utf_8')), ret['out'].split(b'\n'))
            #print(list_remotes)
            return list(list_remotes)[0:-1]
        else:
            raise RCloneException('listremotes returns %d %s' % (ret['code'], ret['error']))

    def ls(self, dest, flags=[]):
        """
        Executes: rclone ls remote:path [flags]
        Args:
        - dest (string): A string "remote:path" representing the location to list.
        """
        return self.run_cmd(command="ls", extra_args=[dest] + flags)

    def lsjson(self, dest, flags=[]):
        """
        Executes: rclone lsjson remote:path [flags]
        Args:
        - dest (string): A string "remote:path" representing the location to list.
        """
        return self.run_cmd(command="lsjson", extra_args=[dest] + flags)

    def delete(self, dest, flags=[]):
        """
        Executes: rclone delete remote:path
        Args:
        - dest (string): A string "remote:path" representing the location to delete.
        """
        return self.run_cmd(command="delete", extra_args=[dest] + flags)

    def check(self, src, dest, flags=[]):
        """
        Executes: rclone check src dest 
        """
        ret = self.run_cmd(command="check", extra_args=[src, dest] + flags)
        if ret['code'] == 0:
            return (0, ret['out'] )
        else:
            raise RCloneException('rclone.check returns error %d (%s)' % (ret['code'], ret['error']))

def with_config(cfgstring):
    """
    Configure a new RClone instance.
    """
    inst = RClone(cfgstring=cfgstring)
    return inst

def test():
    rc = with_config("""[local]
        type = local
        nounc = true""")
    result = rc.listremotes()
    print("result = ", result)

    print("With default cfg file")
    rc = RClone()
    result = rc.listremotes()
    print("result = ", result)

if __name__ == "__main__":
    test()



"""Lineage related commands
"""
import click
from typing import Optional

from dataworkspaces.workspace import Workspace, SnapshotWorkspaceMixin, ResourceRoles
from dataworkspaces.errors import ConfigurationError
from dataworkspaces.utils.lineage_utils import make_simplified_lineage_graph_for_resource


def lineage_graph_command(
    workspace: Workspace,
    output_file: str,
    resource_name: Optional[str],
    snapshot: Optional[str],
    format="html",
    width: int = 1024,
    height: int = 800,
) -> None:
    if not isinstance(workspace, SnapshotWorkspaceMixin):
        raise ConfigurationError(
            "Workspace %s does not support snapshots and lineage" % workspace.name
        )
    if not workspace.supports_lineage():
        raise ConfigurationError("Workspace %s does not support lineage" % workspace.name)
    store = workspace.get_lineage_store()

    snapshot_hash = None  # type: Optional[str]
    if snapshot is not None:
        md = workspace.get_snapshot_by_tag_or_hash(snapshot)
        snapshot_hash = md.hashval
    if resource_name is not None:
        workspace.validate_resource_name(resource_name)
    else:
        for r in workspace.get_resource_names():
            if workspace.get_resource_role(r) == ResourceRoles.RESULTS:
                resource_name = r
                break
        if resource_name is None:
            raise ConfigurationError(
                "Did not find a results resource in workspace. If you want to graph the lineage of a non-results resource, use the --resource option."
            )
    make_simplified_lineage_graph_for_resource(
        workspace.get_instance(),
        store,
        resource_name,
        output_file,
        snapshot_hash=snapshot_hash,
        format=format,
        width=width,
        height=height,
    )
    if snapshot is None:
        click.echo("Wrote lineage for %s to %s" % (resource_name, output_file))
    else:
        click.echo(
            "Wrote lineage for %s as of snapshot %s to %s" % (resource_name, snapshot, output_file)
        )



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
This is an API for selected Data Workspaces management functions.
"""
from typing import Optional, NamedTuple, List, Iterable, cast, Tuple

from dataworkspaces import __version__
from dataworkspaces.workspace import (
    find_and_load_workspace,
    LocalStateResourceMixin,
    SnapshotWorkspaceMixin,
    JSONDict,
    FileResourceMixin
)
from dataworkspaces.commands.snapshot import snapshot_command
from dataworkspaces.commands.restore import restore_command
from dataworkspaces.commands.lineage import lineage_graph_command
from dataworkspaces.commands.report import _get_results
from dataworkspaces.errors import ConfigurationError
import dataworkspaces.utils.lineage_utils as lu

__api_version__ = "0.2"


def get_version():
    """Get the version string for the installed version of Data Workspaces"""
    return __version__


def get_api_version():
    """The API version is maintained independently of the overall DWS version.
    It should be more stable.
    """
    return __api_version__


class ResourceInfo(NamedTuple):
    """Named tuple representing the
    results from a call to :func:`~get_resource_info`.
    """

    name: str
    role: str
    resource_type: str
    local_path: Optional[str]


def get_resource_info(workspace_uri_or_path: Optional[str] = None, verbose: bool = False
    ) -> List[ResourceInfo]:
    """Returns a list of ResourceInfo instances, describing the resources
    defined for this workspace.
    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)

    return [
        ResourceInfo(
            r.name,
            r.role,
            r.resource_type,
            cast(LocalStateResourceMixin, r).get_local_path_if_any()
            if isinstance(r, LocalStateResourceMixin)
            else None,
        )
        for r in workspace.get_resources()
    ]

def get_local_path_for_resource(
        name: str,
        workspace_uri_or_path: Optional[str] = None,
        verbose: bool = False
    ) -> Optional[str]:
    """If a local path is available for this resource, return it. Otherwise,
    return None."""
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    r = workspace.get_resource(name)
    return  cast(LocalStateResourceMixin, r).get_local_path_if_any() \
            if isinstance(r, LocalStateResourceMixin) \
            else None

class ResourceFileSystem:
    """subset of fsspec supported by our file resources."""
    def __init__(self, resource:FileResourceMixin):
        self.resource = resource

    def open(self, rel_path:str, mode:str):
        return self.resource.open(rel_path, mode)

    def ls(self, rel_path:str) -> List[str]:
        return self.resource.ls(rel_path)

    def exists(self, rel_path) -> bool:
        return self.resource.does_subpath_exist(rel_path)

    def isfile(self, rel_path) -> bool:
        return self.resource.does_subpath_exist(rel_path, must_be_file=True)

    def isdir(self, rel_path) -> bool:
        return self.resource.does_subpath_exist(rel_path, must_be_directory=True)


def get_filesystem_for_resource(name: str,
        workspace_uri_or_path: Optional[str] = None,
        verbose: bool = False
    ) -> Optional[ResourceFileSystem]:
    """Get the a filesystem-like object for the named resource.
    If it isn't a FileResource, returns None.
    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    resource = workspace.get_resource(name)
    if isinstance(resource, FileResourceMixin):
        return ResourceFileSystem(resource)
    else:
        return None


class SnapshotInfo(NamedTuple):
    """Named tuple represneting the results from a call
    to :func:`~get_snapshot_history`
    """

    snapshot_number: int
    hashval: str
    tags: List[str]
    timestamp: str
    message: str
    metrics: Optional[JSONDict]


def take_snapshot(
    workspace_uri_or_path: Optional[str] = None,
    tag: Optional[str] = None,
    message: str = "",
    verbose: bool = False,
) -> str:
    """Take a snapshot of the workspace, using the tag and message,
    if provided. Returns the snapshot hash (which can be used to restore to
    this point).
    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    return snapshot_command(workspace, tag=tag, message=message)


def get_snapshot_history(
    workspace_uri_or_path: Optional[str] = None,
    reverse: bool = False,
    max_count: Optional[int] = None,
    verbose: bool = False,
) -> Iterable[SnapshotInfo]:
    """Get the history of snapshots, starting with the oldest first (unless :reverse: is True).
    Returns a list of SnapshotInfo instances, containing the snapshot number,
    hash, tag, timestamp, and message. If :max_count: is specified, returns at most that many snapshots.

    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    assert isinstance(workspace, SnapshotWorkspaceMixin)
    if not reverse:
        return [
            SnapshotInfo(
                snapshot_idx + 1, md.hashval, md.tags, md.timestamp, md.message, md.metrics
            )
            for (snapshot_idx, md) in enumerate(
                workspace.list_snapshots(reverse=False, max_count=max_count)
            )
        ]
    else:
        last_snapshot_no = workspace.get_next_snapshot_number() - 1
        return [
            SnapshotInfo(
                last_snapshot_no - i, md.hashval, md.tags, md.timestamp, md.message, md.metrics
            )
            for (i, md) in enumerate(workspace.list_snapshots(reverse=True, max_count=max_count))
        ]


def restore(
    tag_or_hash: str,
    workspace_uri_or_path: Optional[str] = None,
    only: Optional[List[str]] = None,
    leave: Optional[List[str]] = None,
    verbose: bool = False,
) -> int:
    """Restore to a previous snapshot, identified by either its hash
    or its tag (if one was specified). Parameters:

    * ``only`` - an optional list of resources to store. If specified
      all other resources will be left as-is.
    * ``leave`` - an optional list of resource to leave as-is. Both
      ``only`` and ``leave`` should not be specified together.

    Returns the number of resources changed.
    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    return restore_command(workspace, tag_or_hash=tag_or_hash, only=only, leave=leave)


def make_lineage_table(
    workspace_uri_or_path: Optional[str] = None,
    tag_or_hash: Optional[str] = None,
    verbose: bool = False,
) -> Iterable[Tuple[str, str, str, Optional[List[str]]]]:
    """Make a table of the lineage for each resource.
    The columns are: ref, lineage type, details, inputs
    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    if not isinstance(workspace, SnapshotWorkspaceMixin):
        raise ConfigurationError("Workspace %s does not support lineage" % workspace.name)
    if not workspace.supports_lineage():
        raise ConfigurationError("Workspace %s does not support lineage" % workspace.name)
    snapshot_hash = None  # type: Optional[str]
    if tag_or_hash is not None:
        md = workspace.get_snapshot_by_tag_or_hash(tag_or_hash)
        snapshot_hash = md.hashval
    return lu.make_lineage_table(
        workspace.get_instance(), workspace.get_lineage_store(), snapshot_hash
    )


def make_lineage_graph(
    output_file: str,
    workspace_uri_or_path: Optional[str] = None,
    resource_name: Optional[str] = None,
    tag_or_hash: Optional[str] = None,
    width: int = 1024,
    height: int = 800,
    verbose: bool = False,
) -> None:
    """Write a lineage graph as an html/javascript page to the specified file.
    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    lineage_graph_command(
        workspace,
        output_file,
        resource_name=resource_name,
        snapshot=tag_or_hash,
        width=width,
        height=height,
    )


def get_results(
    workspace_uri_or_path: Optional[str] = None,
    tag_or_hash: Optional[str] = None,
    resource_name: Optional[str] = None,
    verbose: bool = False,
) -> Optional[Tuple[JSONDict, str]]:
    """Get a results file as a parsed json dict. If no resource or snapshot
    is specified, searches all the results resources for a file. If a snapshot
    is specified, we look in the subdirectory where the resuls have been moved.
    If no snapshot is specified, and we don't find a file, we look in the most
    recent snapshot.

    Returns a tuple with the results and the logical path (resource:/subpath) to
    the results. If nothing is found, returns None.
    """
    workspace = find_and_load_workspace(True, verbose, workspace_uri_or_path)
    return _get_results(workspace, tag_or_hash, resource_name)



"""
Utilities related to hashing, particularly for simulating git's hashing
without necessarily requiring git to be installed.
"""

import hashlib
import re

HASH_RE = re.compile(r"^[0-9a-fA-F]+$")


def is_a_git_hash(s):
    return len(s) == 40 and (HASH_RE.match(s) is not None)


MIN_SHORT_HASH_LEN = 6
# short hashes must be lowercase
SHORT_HASH_RE = re.compile(r"^[0-9a-f]+$")


def is_a_shortened_git_hash(s):
    """We can refer to snapshots using the first 6+ characters
    of the hash
    """
    return len(s) >= MIN_SHORT_HASH_LEN and (SHORT_HASH_RE.match(s) is not None)


def hash_file(fpath):
    """Compute the same hash on the file as git would (e.g. via git hash-object).
    The hash is the sha1 digest, but with a header added to the file first:
    the word "blob", followed by a space, followed by the content length,
    followed by a zero byte.
    """
    with open(fpath, "rb") as f:
        data = f.read()
    size = len(data)
    header = ("blob %d" % size).encode("ascii") + b"\0"
    blob = header + data
    m = hashlib.sha1()
    m.update(blob)
    return m.hexdigest()


def hash_bytes(data: bytes):
    """Compute the same hash on bytes as git would (e.g. via git hash-object).
    The hash is the sha1 digest, but with a header added to the bytes first:
    the word "blob", followed by a space, followed by the content length,
    followed by a zero byte.
    """
    assert isinstance(data, bytes)
    size = len(data)
    header = ("blob %d" % size).encode("ascii") + b"\0"
    blob = header + data
    m = hashlib.sha1()
    m.update(blob)
    return m.hexdigest()


def _test(fname):
    hv = hash_file(__file__)
    assert is_a_git_hash(hv), "hv wasn't a git hash"
    short_hash = hv[0:6]
    assert is_a_shortened_git_hash(hv), "%s wasn't a short git hash" % short_hash
    print(hv)


if __name__ == "__main__":
    # testing
    _test(__file__)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Utilities for dealing with snapshots, particularly for resources where
we move files into a unique snapshot subdirectory each time a
snapshot is taken.
"""
import os
from os.path import join, exists
import re
import stat
from tempfile import NamedTemporaryFile
import shutil
from typing import Set, List, Pattern, Union, Callable

import click

from dataworkspaces.errors import ConfigurationError
from dataworkspaces.utils.file_utils import remove_dir_if_empty
from dataworkspaces.utils.hash_utils import hash_file

# Timestamps have the form '2018-09-30T14:09:05'
TEMPLATE_VAR_PATS = {
    "USERNAME": r"\w([\w\-\.])*",
    "HOSTNAME": r"\w([\w\-\.])*",
    "SNAPSHOT_NO": r"\d\d+",
    "ISO_TIMESTAMP": r"\d\d\d\d\-\d\d\-\d\dT\d\d:\d\d:\d\d",
    "YEAR": r"\d\d\d\d",
    "MONTH": r"\d\d",
    "SHORT_MONTH": r"\w\w\w",
    "DAY": r"\d\d",
    "HOUR": r"\d\d",
    "MIN": r"\d\d",
    "SEC": r"\d\d",
    "DAY_OF_WEEK": r"\w+",
    "TAG": r"\w([\w\-\.])*",
    #'TAGBOTH':r'\-(\w+\-)?',
    #'TAGLEFT':r'(\-\w+)?',
    #'TAGRIGHT':r'(\w+\-)?'
}
TEMPLATE_VAR_RE = re.compile("\\{\w+\\}")

VALID_TEMPLATE_VARS = set(TEMPLATE_VAR_PATS.keys())
# remove the 'internal' template variables
# VALID_TEMPLATE_VARS.remove('TAGBOTH')
# VALID_TEMPLATE_VARS.remove('TAGLEFT')
# VALID_TEMPLATE_VARS.remove('TAGRIGHT')


def validate_template(template):
    if not template.startswith("snapshots/"):
        raise ConfigurationError("Templates must start with 'snapshots/'")
    for mo in TEMPLATE_VAR_RE.finditer(template):
        tvar = mo.group(0)[1:-1]
        if tvar not in VALID_TEMPLATE_VARS:
            raise ConfigurationError(
                "Unknown variable '%s' in results directory template '%s'" % (tvar, template)
            )


# def _translate_tag_vars(template):
#     """The {TAG} template variable is optional, so we treat it as a
#     special case to avoid generated paths like foo/bar- or fooo/bar--baz
#     when the tag is not present. We replace any occurrances of a {TAG}
#     that has dashes on either side with one a template variable that
#     indicates where the dashes are.
#     """
#     return template.replace('-{TAG}-', '{TAGBOTH}')\
#                    .replace('-{TAG}', '{TAGLEFT}')\
#                    .replace('{TAG}-', '{TAGRIGHT}')


def make_re_pattern_for_dir_template(template):
    # template = _translate_tag_vars(template)
    # we escape the template and then "unescape" the braces
    escaped = re.escape(template).replace("\\{", "{").replace("\\}", "}")

    def repl(tvar_mo):
        tvar = tvar_mo.group(0)[1:-1]
        assert tvar in TEMPLATE_VAR_PATS, "Uknown result directory template variable '%s'" % tvar
        return TEMPLATE_VAR_PATS[tvar]

    return "^" + TEMPLATE_VAR_RE.sub(repl, escaped) + "$"


WEEKDAYS = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
SHORT_MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]


def expand_dir_template(template, username, hostname, timestamp, snapshot_no, snapshot_tag=None):
    # template = _translate_tag_vars(template)
    values = {
        "USERNAME": username,
        "HOSTNAME": hostname,
        "ISO_TIMESTAMP": timestamp.isoformat()[0:19],  # truncate fractional seconds
        "YEAR": "%04d" % timestamp.year,
        "MONTH": "%02d" % timestamp.month,
        "SHORT_MONTH": SHORT_MONTHS[timestamp.month - 1],
        "DAY": "%02d" % timestamp.day,
        "HOUR": "%02d" % timestamp.hour,
        "MIN": "%02d" % timestamp.minute,
        "SEC": "%02d" % timestamp.second,
        "DAY_OF_WEEK": WEEKDAYS[timestamp.weekday()],
        "SNAPSHOT_NO": "%03d" % snapshot_no,
    }
    if snapshot_tag is not None:
        values["TAG"] = snapshot_tag
        # values['TAGBOTH'] = '-' + snapshot_tag + '-'
        # values['TAGLEFT'] = '-' + snapshot_tag
        # values['TAGRIGHT'] = snapshot_tag + '-'
    else:
        # values['TAG'] = values['TAGLEFT'] = values['TAGRIGHT'] = ''
        # values['TAGBOTH'] = '-'
        values["TAG"] = "%03d" % snapshot_no  # if the tag isn't present, use the snapshot number

    def repl(tvar_mo):
        tvar = tvar_mo.group(0)[1:-1]
        assert tvar in values, "Uknown result directory template variable '%s'" % tvar
        return values[tvar]

    return TEMPLATE_VAR_RE.sub(repl, template)


def move_file_and_set_readonly(src: str, dest: str) -> None:
    os.rename(src, dest)
    mode = os.stat(dest).st_mode
    os.chmod(dest, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)


def copy_file_and_set_readonly(src: str, dest: str) -> None:
    shutil.copyfile(src, dest)
    mode = os.stat(dest).st_mode
    os.chmod(dest, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)


DOT_GIT_RE = re.compile("^" + re.escape(".git") + "$")


def move_current_files_local_fs(
    resource_name: str,
    base_dir: str,
    rel_dest_root: str,
    exclude_files: Set[str],
    exclude_dirs_res: Union[Pattern, List[Pattern]],
    move_fn: Callable[[str, str], None] = move_file_and_set_readonly,
    verbose: bool = False,
):
    """Utility for impelementing Resource.results_move_current_file()
    for when the files are stored on the local filesystem (e.g. local or git
    resources).
    exclude_dirs_res is either a single regular expression object or a list
    of regular expression objects. It should not include .git, as that will always
    be appended to the list.

    The move function should also set file to read-only. For git, this should
    be done before adding the file to the staging area.

    This returns the list of (before, after) relative path pairs.
    """
    abs_dest_root = join(base_dir, rel_dest_root)
    created_dir = False  # only create when we actually have a file to move
    moved_files = []
    dirs_to_remove_if_empty = []  # type: List[str]
    exclude_dirs_re_list = (
        exclude_dirs_res if isinstance(exclude_dirs_res, list) else [exclude_dirs_res,]
    )  # type: List[Pattern]
    exclude_dirs_re_list.append(DOT_GIT_RE)
    for (dirpath, dirnames, filenames) in os.walk(base_dir):
        assert dirpath.startswith(base_dir)
        rel_dirpath = dirpath[len(base_dir) + 1 :]

        def join_to_rel_dirpath(f):
            return join(rel_dirpath, f)

        join_rel_path = join_to_rel_dirpath if len(rel_dirpath) > 0 else lambda f: f
        # find directories we should skip, as they represent results from
        # prior runs
        skip = []
        for (i, dir_name) in enumerate(dirnames):
            abs_dirpath = join(dirpath, dir_name)
            rel_dirpath = abs_dirpath[len(base_dir) + 1 :]
            for exclude_dirs_re in exclude_dirs_re_list:
                if exclude_dirs_re.match(rel_dirpath):
                    skip.append(i)
                    if verbose:
                        print("Skipping directory %s" % rel_dirpath)
        skip.reverse()
        for i in skip:
            del dirnames[i]
        # move files to our new directory
        moved_files_out_of_this_dir = False
        for f in filenames:
            rel_src_file = join_rel_path(f)
            if rel_src_file in exclude_files:
                if verbose:
                    click.echo("[%s] Leaving %s" % (resource_name, rel_src_file))
                continue
            assert not rel_src_file.startswith("snapshot/"), (
                "Internal error: file %s should have been excluded from move" % rel_src_file
            )
            rel_dest_file = join(rel_dest_root, join_rel_path(f))
            abs_src_file = join(dirpath, f)
            abs_dest_file = join(abs_dest_root, join_rel_path(f))
            if verbose:
                click.echo("[%s] Moving %s to %s" % (resource_name, rel_src_file, rel_dest_file))
                click.echo("     Absolute %s => %s" % (abs_src_file, abs_dest_file))
            dest_parent = os.path.dirname(abs_dest_file)
            if not created_dir:
                # lazily create the root directory
                os.makedirs(abs_dest_root)
                created_dir = True
            if not exists(dest_parent):
                os.makedirs(dest_parent)
            move_fn(abs_src_file, abs_dest_file)
            moved_files.append((rel_src_file, rel_dest_file))
            moved_files_out_of_this_dir = True
        if moved_files_out_of_this_dir:
            dirs_to_remove_if_empty.append(dirpath)
    for dirpath in dirs_to_remove_if_empty:
        remove_dir_if_empty(dirpath, base_dir, verbose=verbose)
    return moved_files


def copy_current_files_local_fs(
    resource_name: str,
    base_dir: str,
    rel_dest_root: str,
    exclude_files: Set[str],
    exclude_dirs_res: Union[Pattern, List[Pattern]],
    copy_fn: Callable[[str, str], None] = copy_file_and_set_readonly,
    verbose: bool = False,
) -> List[str]:
    """Utility for impelementing Resource.results_copy_current_file()
    for when the files are stored on the local filesystem (e.g. local or git
    resources).
    exclude_dirs_res is either a single regular expression object or a list
    of regular expression objects. It should not include .git, as that will always
    be appended to the list.

    The copy function should also set file to read-only. For git, this should
    be done before adding the file to the staging area.

    This returns the list of the copied relative path pairs.
    """
    abs_dest_root = join(base_dir, rel_dest_root)
    created_dir = False  # only create when we actually have a file to move
    copied_files = []  # type: List[str]
    exclude_dirs_re_list = (
        exclude_dirs_res if isinstance(exclude_dirs_res, list) else [exclude_dirs_res,]
    )  # type: List[Pattern]
    exclude_dirs_re_list.append(DOT_GIT_RE)
    for (dirpath, dirnames, filenames) in os.walk(base_dir):
        assert dirpath.startswith(base_dir)
        rel_dirpath = dirpath[len(base_dir) + 1 :]

        def join_to_rel_dirpath(f):
            return join(rel_dirpath, f)

        join_rel_path = join_to_rel_dirpath if len(rel_dirpath) > 0 else lambda f: f
        # find directories we should skip, as they represent results from
        # prior runs
        skip = []
        for (i, dir_name) in enumerate(dirnames):
            abs_dirpath = join(dirpath, dir_name)
            rel_dirpath = abs_dirpath[len(base_dir) + 1 :]
            for exclude_dirs_re in exclude_dirs_re_list:
                if exclude_dirs_re.match(rel_dirpath):
                    skip.append(i)
                    if verbose:
                        print("Skipping directory %s" % rel_dirpath)
        skip.reverse()
        for i in skip:
            del dirnames[i]
        # copy files to our new directory
        for f in filenames:
            rel_src_file = join_rel_path(f)
            if rel_src_file in exclude_files:
                if verbose:
                    click.echo("[%s] Leaving %s" % (resource_name, rel_src_file))
                continue
            assert not rel_src_file.startswith("snapshot/"), (
                "Internal error: file %s should have been excluded from move" % rel_src_file
            )
            rel_dest_file = join(rel_dest_root, join_rel_path(f))
            abs_src_file = join(dirpath, f)
            abs_dest_file = join(abs_dest_root, join_rel_path(f))
            if verbose:
                click.echo("[%s] Copying %s to %s" % (resource_name, rel_src_file, rel_dest_file))
                click.echo("     Absolute %s => %s" % (abs_src_file, abs_dest_file))
            dest_parent = os.path.dirname(abs_dest_file)
            if not created_dir:
                # lazily create the root directory
                os.makedirs(abs_dest_root)
                created_dir = True
            if not exists(dest_parent):
                os.makedirs(dest_parent)
            copy_fn(abs_src_file, abs_dest_file)
            copied_files.append(rel_dest_file)
    return copied_files


def write_and_hash_file(write_fn, filename_template, verbose):
    """Write a file to a temporary location, take its hash
    and rename to filename_template, replacing <HASHVAL> with
    the hash. Returns hashval, the snapshot filename, and
    a boolean indicating whether this is a new snapshot
    (may end up with a snapshot matching a previous one).
    """
    with NamedTemporaryFile(delete=False) as f:
        tfilename = f.name
    try:
        write_fn(tfilename)
        hashval = hash_file(tfilename)
        target_name = filename_template.replace("<HASHVAL>", hashval)
        assert target_name != filename_template
        if exists(target_name):
            return (hashval, target_name, False)
        else:
            # issue #21 - need to use a copy instead of a rename,
            # in case /tmp is on a different filesystem.
            shutil.copyfile(tfilename, target_name)
            return (hashval, target_name, True)
    finally:
        if exists(tfilename):
            os.remove(tfilename)



"""Utilities for git repos that use git-lfs for support of large
files.
"""
import os
from os.path import join, exists
from typing import Optional
from pathlib import Path

import click

from dataworkspaces.utils.git_utils import is_git_repo, git_add, git_commit
from dataworkspaces.utils.subprocess_utils import find_exe, call_subprocess


def _does_attributes_file_reference_lfs(fpath) -> bool:
    with open(fpath, "r") as f:
        for line in f:
            words = line.split()
            if len(words) < 2:
                continue
            # start with index 1 to skip the pattern
            for attr in words[1:]:
                if attr.endswith("=lfs"):
                    return True
        return False


def is_a_git_lfs_repo(repo_dir: str, recursive: bool = True) -> bool:
    assert is_git_repo(repo_dir), "%s is not a git repo" % repo_dir
    if recursive:
        for (dirpath, filenames, dirnames) in os.walk(repo_dir):
            fname = join(dirpath, ".gitattributes")
            if exists(fname):
                if _does_attributes_file_reference_lfs(fname):
                    return True
        return False
    else:  # just check the root
        fname = join(repo_dir, ".gitattributes")
        if exists(fname):
            return _does_attributes_file_reference_lfs(fname)
        else:
            return False


def is_git_lfs_installed_for_user(home_dir: str = str(Path.home())) -> bool:
    git_config = join(home_dir, ".gitconfig")
    if not exists(git_config):
        return False
    with open(git_config, "r") as f:
        for line in f:
            if line.strip() == '[filter "lfs"]':
                return True
        return False


# Utility funtions for issue #12 - if a repo is git-lfs enabled, and git-lfs is not in the path,
# git add will fail silently for filter calls (e.g. in git add). We explicitly check that
# the executable is in the path in situations where we will call git as a subprocess.

GIT_LFS_ERRMSG = "git-lfs does not seem to be installed on your system. Install it or, if it is already installed, make sure that it is in your PATH."


def find_git_lfs_in_path() -> str:
    """Validate that git-lfs is in the path, asssuming we already know that this
    is a git-lfs repo.
    If the executable is not found, throw a configuration error. We need to do this, as git itself
    will not return an error return code if a filter (e.g. git-lfs) is not found.

    Retuns the path to the git-lfs executable.
    """
    return find_exe("git-lfs", GIT_LFS_ERRMSG, additional_search_locations=[])


def ensure_git_lfs_installed_for_user(lfs_exe, verbose: bool = False) -> bool:
    """Run the install operation if necessary. Returns true if it was necessary,
    false otherwise.
    """
    if not is_git_lfs_installed_for_user():
        click.echo("Git lfs not installed for your account, installing...")
        call_subprocess([lfs_exe, "install"], cwd=str(Path.home()), verbose=verbose)
        return True
    elif verbose:
        click.echo("git-lfs is installed for this user")
    return False


def init_git_lfs(
    workspace_dir: str, git_lfs_attributes: Optional[str] = None, verbose: bool = False
):
    """Called during the dws init if the repo has references to lfs in its
    .gitattributes files or if the user requested lfs file wildcares via
    git_lfs_attributes
    """
    lfs_exe = find_git_lfs_in_path()
    ensure_git_lfs_installed_for_user(lfs_exe, verbose=verbose)
    if git_lfs_attributes:
        for extn in git_lfs_attributes.split(","):
            call_subprocess([lfs_exe, "track", extn], cwd=workspace_dir, verbose=verbose)
        git_add(workspace_dir, [".gitattributes"], verbose=verbose)
        git_commit(workspace_dir, "added git-lfs attributes", verbose=verbose)


def ensure_git_lfs_configured_if_needed(repo_dir: str, verbose: bool = False) -> None:
    """If this repo uses git-lfs, then 1) validate that git-lfs is in the path,
    and 2) run git-lfs install for the user, if needed.
    If the repo uses git-lfs, but we cannot find the executable,
    throw a configuration error. We need to do this, as git itself
    will not return an error return code if a filter (e.g. git-lfs) is not found.
    """
    if not is_a_git_lfs_repo(repo_dir, recursive=True):
        return
    lfs_exe = find_exe("git-lfs", GIT_LFS_ERRMSG, additional_search_locations=[])
    need_to_download = ensure_git_lfs_installed_for_user(lfs_exe, verbose=verbose)
    if need_to_download:
        # If the user wasn't configured for git-lfs when cloning, we need to
        # explicitly download the files.
        call_subprocess([lfs_exe, "fetch"], cwd=repo_dir, verbose=verbose)
        call_subprocess([lfs_exe, "checkout"], cwd=repo_dir, verbose=verbose)



"""
Git backend for storing a workspace
"""

import os
from os.path import (
    exists,
    join,
    isdir,
    basename,
    isabs,
    abspath,
    expanduser,
    dirname,
    curdir,
    commonpath,
)
import shutil
import json
import re
import uuid
from urllib.parse import ParseResult, urlparse
from typing import Any, Iterable, Optional, List, Dict, Tuple, cast

assert Dict  # make pyflakes happy

import click

import dataworkspaces.workspace as ws
from dataworkspaces.workspace import JSONDict, SnapshotMetadata
from dataworkspaces.errors import ConfigurationError, InternalError
from dataworkspaces.utils.subprocess_utils import call_subprocess
from dataworkspaces.utils.git_utils import (
    commit_changes_in_repo,
    git_init,
    git_add,
    is_git_dirty,
    is_pull_needed_from_remote,
    GIT_EXE_PATH,
    set_remote_origin,
    verify_git_config_initialized,
    git_remove_file,
    git_remove_subtree,
    get_branch_info,
    ensure_entry_in_gitignore,
    echo_git_status_for_user,
    switch_git_branch_if_needed
)
from dataworkspaces.utils.git_fat_utils import (
    validate_git_fat_in_path_if_needed,
    run_git_fat_pull_if_needed,
    validate_git_fat_in_path,
    run_git_fat_push_if_needed,
    setup_git_fat_for_repo,
    is_a_git_fat_repo,
)
from dataworkspaces.utils.git_lfs_utils import (
    init_git_lfs,
    is_a_git_lfs_repo,
    ensure_git_lfs_configured_if_needed,
)
from dataworkspaces.utils.file_utils import safe_rename, get_subpath_from_absolute
from dataworkspaces.utils.param_utils import (
    HOSTNAME,
    init_scratch_directory,
    clone_scratch_directory,
    get_scratch_directory,
    SCRATCH_DIRECTORY,
    LOCAL_SCRATCH_DIRECTORY,
    DWS_GIT_BRANCH
)
from dataworkspaces.utils.lineage_utils import (
    FileLineageStore,
    LineageStore,
    ResourceRef,
    ResourceLineage,
)


BASE_DIR = ".dataworkspace"
GIT_IGNORE_FILE_PATH = ".dataworkspace/.gitignore"
CONFIG_FILE_PATH = ".dataworkspace/config.json"
LOCAL_PARAMS_PATH = ".dataworkspace/local_params.json"
RESOURCES_FILE_PATH = ".dataworkspace/resources.json"
RESOURCE_LOCAL_PARAMS_PATH = ".dataworkspace/resource_local_params.json"
SNAPSHOT_DIR_PATH = ".dataworkspace/snapshots"
SNAPSHOT_METADATA_DIR_PATH = ".dataworkspace/snapshot_metadata"
CURRENT_LINEAGE_DIR_PATH = ".dataworkspace/current_lineage"
SNAPSHOT_LINEAGE_DIR_PATH = ".dataworkspace/snapshot_lineage"


class GitFileLineageStore(FileLineageStore):
    """Subclass of file lineage store that adds
    the lineage files to the git repo.
    """

    def __init__(self, workspace: "Workspace"):
        super().__init__(
            cast(Workspace, workspace).get_instance(),
            join(workspace.workspace_dir, CURRENT_LINEAGE_DIR_PATH),
            join(workspace.workspace_dir, SNAPSHOT_LINEAGE_DIR_PATH),
        )
        self.workspace = workspace

    def _add_to_git(self, path: str):
        ws_dir = cast(str, self.workspace.workspace_dir)
        git_add(
            ws_dir,
            [get_subpath_from_absolute(ws_dir, path)],  # type: ignore
            verbose=self.workspace.verbose,
        )

    def _save_rfile_to_snapshot(
        self,
        resource_name: str,
        lineage_map: Dict[ResourceRef, ResourceLineage],
        snapshot_hash: str,
    ) -> str:
        rpath = super()._save_rfile_to_snapshot(resource_name, lineage_map, snapshot_hash)
        self._add_to_git(rpath)
        return rpath

    def _copy_rfile_to_snapshot(self, resource_name: str, snapshot_hash: str) -> Tuple[str, str]:
        (spath, dpath) = super()._copy_rfile_to_snapshot(resource_name, snapshot_hash)
        self._add_to_git(dpath)
        return (spath, dpath)

    def _write_placeholder_to_snapshot(
        self, snapshot_hash: str, filename: str, content: str
    ) -> str:
        spath = super()._write_placeholder_to_snapshot(snapshot_hash, filename, content)
        self._add_to_git(spath)
        return spath

    def delete_snapshot_lineage(self, instance: str, snapshot_hash: str) -> None:
        """Delete any lineage data associated with the specified snapshot.
        """
        lineage_relative_path = join(SNAPSHOT_LINEAGE_DIR_PATH, snapshot_hash)
        git_remove_subtree(
            self.workspace.workspace_dir, lineage_relative_path, verbose=self.workspace.verbose
        )


class Workspace(ws.Workspace, ws.SyncedWorkspaceMixin, ws.SnapshotWorkspaceMixin):
    def __init__(self, workspace_dir: str, batch: bool = False, verbose: bool = False):
        self.workspace_dir = workspace_dir  # type: str
        cf_data = self._load_json_file(CONFIG_FILE_PATH)
        super().__init__(cf_data["name"], cf_data["dws-version"], batch, verbose)
        self.global_params = cf_data["global_params"]
        self.local_params = self._load_json_file(LOCAL_PARAMS_PATH)
        hostname = self.local_params[
            HOSTNAME
        ]  # TODO: make this user settable when creating the workspace
        assert isinstance(hostname, str)
        self.instance = hostname
        self.resource_params = self._load_json_file(RESOURCES_FILE_PATH)  # type: List[JSONDict]
        self.resource_params_by_name = {}  # type: Dict[str, JSONDict]
        for r in self.resource_params:
            self.resource_params_by_name[r["name"]] = r
        self.resource_local_params_by_name = self._load_json_file(
            RESOURCE_LOCAL_PARAMS_PATH
        )  # type: Dict[str,JSONDict]
        self.lineage_store = GitFileLineageStore(self)
        self.scratch_dir = get_scratch_directory(
            self.workspace_dir, self.global_params, self.local_params
        )

    def get_instance(self) -> str:
        return self.instance

    def supports_lineage(self) -> bool:
        return True

    def get_lineage_store(self) -> LineageStore:
        return self.lineage_store

    def get_scratch_directory(self) -> str:
        if self.scratch_dir is not None:
            return self.scratch_dir
        else:
            raise ConfigurationError(
                "Neither the %s nor %s parameters are set, so cannot find scratch directory. Please set one using 'dws config'."
                % (SCRATCH_DIRECTORY, LOCAL_SCRATCH_DIRECTORY)
            )

    def _load_json_file(self, relative_path):
        f_path = join(self.workspace_dir, relative_path)
        if not exists(f_path):
            raise ConfigurationError("Did not find workspace metadata file %s" % f_path)
        with open(f_path, "r") as f:
            return json.load(f)

    def _save_json_to_file(self, obj, relative_path):
        f_path = join(self.workspace_dir, relative_path)
        with open(f_path, "w") as f:
            json.dump(obj, f, indent=2)

    def _get_global_params(self) -> JSONDict:
        """Get a dict of configuration parameters for this workspace,
        which apply across all instances.
        """
        return self.global_params

    def _get_local_params(self) -> JSONDict:
        """Get a dict of configuration parameters for this particular
        install of the workspace (e.g. local filesystem paths, hostname).
        """
        return self.local_params

    def _set_global_param(self, name: str, value: Any) -> None:
        """Setting does not necessarily take effect until save() is called"""
        data = self._get_global_params()
        data[name] = value
        self._save_json_to_file(
            {"name": self.name, "dws-version": self.dws_version, "global_params": data},
            CONFIG_FILE_PATH,
        )

    def _set_local_param(self, name: str, value: Any) -> None:
        data = self._get_local_params()
        data[name] = value
        self._save_json_to_file(data, LOCAL_PARAMS_PATH)

    def get_resource_names(self) -> Iterable[str]:
        return self.resource_params_by_name.keys()

    def _get_resource_params(self, resource_name) -> JSONDict:
        """Get the parameters for this resource from the workspace's
        metadata store - used when instantitating resources. Show
        throw an exception if resource does not exist.
        """
        if resource_name not in self.resource_params_by_name:
            raise ConfigurationError(
                "A resource by the name '%s' does not exist in this workspace" % resource_name
            )
        return self.resource_params_by_name[resource_name]

    def _get_resource_local_params(self, resource_name: str) -> Optional[JSONDict]:
        """If a resource has local parameters defined for it, return them.
        Otherwise, return None.
        """
        if resource_name in self.resource_local_params_by_name:
            return self.resource_local_params_by_name[resource_name]
        else:
            return None

    def _add_params_for_resource(self, resource_name: str, params: JSONDict) -> None:
        """
        Add the necessary state for a new resource to the workspace.
        """
        assert params["name"] == resource_name
        self.resource_params.append(params)
        self.resource_params_by_name[resource_name] = params
        self._save_json_to_file(self.resource_params, RESOURCES_FILE_PATH)

    def _add_local_params_for_resource(self, resource_name: str, local_params: JSONDict) -> None:
        """
        Add local params either for a new or cloned resource.
        """
        self.resource_local_params_by_name[resource_name] = local_params
        self._save_json_to_file(self.resource_local_params_by_name, RESOURCE_LOCAL_PARAMS_PATH)

    def _set_global_param_for_resource(self, resource_name: str, name: str, value: Any) -> None:
        """It is up to the caller to verify that the resource exists and has
        this parameter defined. Value should be json-serializable (via the to_json() method
        of the param type). Setting does not necessarily take effect until save() is called"""
        assert resource_name in self.resource_params_by_name, (
            "Missing resource params entry for resource %s" % resource_name
        )
        self.resource_params_by_name[resource_name][name] = value
        for pdict in self.resource_params:
            if pdict["name"] == resource_name:
                pdict[name] = value
                self._save_json_to_file(self.resource_params, RESOURCES_FILE_PATH)
                return
        assert 0, "Did not find resource params entry"

    def _set_local_param_for_resource(self, resource_name: str, name: str, value: Any) -> None:
        """It is up to the caller to verify that the resource exists and has
        this parameter defined. Value should be json-serializable (via the to_json() method
        of the param type). Setting does not necessarily take effect until save() is called"""
        assert resource_name in self.resource_local_params_by_name, (
            "Missing resource local params entry for resource %s" % resource_name
        )
        self.resource_local_params_by_name[resource_name][name] = value
        self._save_json_to_file(self.resource_local_params_by_name, RESOURCE_LOCAL_PARAMS_PATH)

    def get_workspace_local_path_if_any(self) -> Optional[str]:
        return self.workspace_dir

    def _add_local_dir_to_gitignore_if_needed(self, resource):
        """Figure out whether resource has a local path under the workspace's
        git repo, which needs to be added to .gitignore. If so, do it.
        """
        if resource.resource_type == "git-subdirectory":
            return  # this is always a part of the dataworkspace's repo
        elif not isinstance(resource, ws.LocalStateResourceMixin):
            return  # no local state, so not an iddue
        local_path = resource.get_local_path_if_any()
        if local_path is None:
            return
        assert isabs(local_path), "Resource local path should be absolute"
        if commonpath([local_path, self.workspace_dir]) != self.workspace_dir:
            return None
        local_relpath = local_path[len(self.workspace_dir) + 1 :]
        if not local_relpath.endswith("/"):
            local_relpath = local_relpath + "/"  # matches only directories
        # Add a / as the start to indicate that the path starts at the root of the repo.
        # Otherwise, we'll hit cases where the path could match other directories (e.g. issue #11)
        local_relpath = "/" + local_relpath if not local_relpath.startswith("/") else local_relpath
        ensure_entry_in_gitignore(
            self.workspace_dir,
            ".gitignore",
            local_relpath,
            match_independent_of_slashes=True,
            verbose=self.verbose,
        )

    def add_resource(
        self, name: str, resource_type: str, role: str, *args, **kwargs
    ) -> ws.Resource:
        r = super().add_resource(name, resource_type, role, *args, **kwargs)
        self._add_local_dir_to_gitignore_if_needed(r)
        return r

    def clone_resource(self, name: str) -> ws.LocalStateResourceMixin:
        """Only called if the resource has local state....
        """
        r = super().clone_resource(name)
        self._add_local_dir_to_gitignore_if_needed(r)
        return r

    def _get_local_scratch_space_for_resource(
        self, resource_name: str, create_if_not_present: bool = False
    ) -> str:
        scratch_path = join(self.workspace_dir, ".dataworkspace/scratch/%s" % resource_name)
        if not isdir(scratch_path):
            if create_if_not_present is False:
                raise InternalError(
                    "Scratch path '%s' for resource %s is missing" % (scratch_path, resource_name)
                )
            os.makedirs(scratch_path)
            ensure_entry_in_gitignore(
                self.workspace_dir,
                ".dataworkspace/.gitignore",
                "/scratch/%s/" % resource_name,
                commit=True,
            )
        return scratch_path

    def save(self, message: str) -> None:
        """Save the current state of the workspace"""
        commit_changes_in_repo(self.workspace_dir, message, verbose=self.verbose)

    def pull_workspace(self) -> ws.SyncedWorkspaceMixin:
        # first, check for problems
        if is_git_dirty(self.workspace_dir):
            raise ConfigurationError(
                "Data workspace metadata repo at %s has uncommitted changes. Please commit before pulling."
                % self.workspace_dir
            )
        validate_git_fat_in_path_if_needed(self.workspace_dir)

        # do the pulls
        branch = self.get_dws_git_branch()
        call_subprocess(
            [GIT_EXE_PATH, "pull", "origin", branch], cwd=self.workspace_dir, verbose=self.verbose
        )
        run_git_fat_pull_if_needed(self.workspace_dir, self.verbose)

        # reload and return new workspace
        return Workspace(self.workspace_dir, batch=self.batch, verbose=self.verbose)

    def _push_precheck(self, resource_list: List[ws.LocalStateResourceMixin]) -> None:
        branch = self.get_dws_git_branch()
        if is_pull_needed_from_remote(self.workspace_dir, branch, self.verbose):
            raise ConfigurationError(
                "Data workspace at %s requires a pull from remote origin" % self.workspace_dir
            )
        validate_git_fat_in_path_if_needed(self.workspace_dir)
        if is_git_dirty(self.workspace_dir):
            if not self.batch:
                click.echo(
                    "Database workspace metadata repo at %s has uncommitted changes:"
                    % self.workspace_dir
                )
                echo_git_status_for_user(self.workspace_dir)
                click.confirm("Do you wish to continue anyway?", abort=True)
            else:
                raise ConfigurationError(
                    "Data workspace metadata repo at %s has uncommitted changes. Please commit before pushing."
                    % self.workspace_dir
                )
        super()._push_precheck(resource_list)

    def push(self, resource_list: List[ws.LocalStateResourceMixin]) -> None:
        branch = self.get_dws_git_branch()
        super().push(resource_list)
        call_subprocess(
            [GIT_EXE_PATH, "push", "origin", branch], cwd=self.workspace_dir, verbose=self.verbose
        )
        run_git_fat_push_if_needed(self.workspace_dir, verbose=self.verbose)

    def publish(self, *args) -> None:
        if len(args) != 1:
            raise InternalError("publish takes one argument: remote_repository, got %s" % args)
        set_remote_origin(self.workspace_dir, args[0], verbose=self.verbose)

    def get_next_snapshot_number(self) -> int:
        """Snapshot numbers are assigned based on how many snapshots have
        already been taken. Counting starts at 1. Note that snaphsot
        numbers are not necessarily unique, as people could simultaneously
        take snapshots in different copies of the workspace. Thus, we
        usually combine the snapshot with the hostname.
        """
        md_dirpath = join(self.workspace_dir, SNAPSHOT_METADATA_DIR_PATH)
        if not isdir(md_dirpath):
            return 1  # first snapshot
        # we recursively walk the tree to be future-proof in case we
        # find that we need to start putting metadata into subdirectories.
        def process_dir(dirpath):
            cnt = 0
            for f in os.listdir(dirpath):
                p = join(dirpath, f)
                if isdir(p):
                    cnt += process_dir(p)
                elif f.endswith("_md.json"):
                    cnt += 1
            return cnt

        return 1 + process_dir(md_dirpath)

    def get_snapshot_metadata(self, hash_val: str) -> SnapshotMetadata:
        hash_val = hash_val.lower()
        md_filename = join(
            join(self.workspace_dir, SNAPSHOT_METADATA_DIR_PATH), "%s_md.json" % hash_val
        )
        if not exists(md_filename):
            raise ConfigurationError("No metadata entry for snapshot %s" % hash_val)
        with open(md_filename, "r") as f:
            data = json.load(f)
        md = ws.SnapshotMetadata.from_json(data)
        assert md.hashval == hash_val
        return md

    def get_snapshot_by_tag(self, tag: str) -> SnapshotMetadata:
        """Given a tag, return the asssociated snapshot metadata.
        This lookup could be slower ,if a reverse index is not kept."""
        md_dir = join(self.workspace_dir, SNAPSHOT_METADATA_DIR_PATH)
        if not exists(md_dir):
            raise ConfigurationError(f"Snapshot for tag {tag} not found")
        regexp = re.compile(re.escape(tag))
        for fname in os.listdir(md_dir):
            if not fname.endswith("_md.json"):
                continue
            fpath = join(md_dir, fname)
            with open(fpath, "r") as f:
                raw_data = f.read()
            if regexp.search(raw_data) is not None:
                md = SnapshotMetadata.from_json(json.loads(raw_data))
                if md.has_tag(tag):
                    return md
        raise ConfigurationError("Snapshot for tag %s not found" % tag)

    def get_snapshot_by_partial_hash(self, partial_hash: str) -> SnapshotMetadata:
        """Given a partial hash for the snapshot, find the snapshot whose hash
        starts with this prefix and return the metadata
        asssociated with the snapshot.
        """
        partial_hash = partial_hash.lower()
        md_dir = join(self.workspace_dir, SNAPSHOT_METADATA_DIR_PATH)
        for fname in os.listdir(md_dir):
            if not fname.endswith("_md.json"):
                continue
            hashval = fname[0:-8].lower()
            if not hashval.startswith(partial_hash):
                continue
            return self.get_snapshot_metadata(hashval)
        raise ConfigurationError("Snapshot match for partial hash %s not found" % partial_hash)

    def _get_snapshot_manifest_as_bytes(self, hash_val: str) -> bytes:
        snapshot_dir = join(self.workspace_dir, SNAPSHOT_DIR_PATH)
        snapshot_file = join(snapshot_dir, "snapshot-%s.json" % hash_val.lower())
        if not exists(snapshot_file):
            raise ConfigurationError("No snapshot found for hash value %s" % hash_val)
        with open(snapshot_file, "rb") as f:
            return f.read()

    def list_snapshots(
        self, reverse: bool = True, max_count: Optional[int] = None
    ) -> Iterable[SnapshotMetadata]:
        """Returns an iterable of snapshot metadata, sorted by timestamp ascending
        (or descending if reverse is True). If max_count is specified, return at
        most that many snaphsots.
        """
        md_dir = join(self.workspace_dir, SNAPSHOT_METADATA_DIR_PATH)
        snapshots = []
        for fname in os.listdir(md_dir):
            if not fname.endswith("_md.json"):
                continue
            with open(join(md_dir, fname), "r") as f:
                snapshots.append(SnapshotMetadata.from_json(json.load(f)))
        snapshots.sort(key=lambda md: md.timestamp, reverse=reverse)
        return snapshots if max_count is None else snapshots[0:max_count]

    def _delete_snapshot_metadata_and_manifest(self, hash_val: str) -> None:
        """Given a snapshot hash, delete the associated metadata.
        """
        rel_snapshot_file = join(SNAPSHOT_DIR_PATH, "snapshot-%s.json" % hash_val.lower())
        git_remove_file(self.workspace_dir, rel_snapshot_file, verbose=self.verbose)
        rel_metadata_file = join(SNAPSHOT_METADATA_DIR_PATH, "%s_md.json" % hash_val.lower())
        git_remove_file(self.workspace_dir, rel_metadata_file, verbose=self.verbose)

    def _snapshot_precheck(self, current_resources: Iterable[ws.Resource]) -> None:
        """Run any prechecks before taking a snapshot. This should throw
        a ConfigurationError if the snapshot would fail for some reason.
        """
        # call prechecks on the individual resources
        super()._snapshot_precheck(current_resources)
        validate_git_fat_in_path_if_needed(self.workspace_dir)

    def _restore_precheck(
        self,
        restore_hashes: Dict[str, Optional[str]],
        restore_resources: List[ws.SnapshotResourceMixin],
    ) -> None:
        """Run any prechecks before restoring. This should throw
        a ConfigurationError if the restore would fail for some reason.
        """
        # call prechecks on the individual resources
        super()._restore_precheck(restore_hashes, restore_resources)
        validate_git_fat_in_path_if_needed(self.workspace_dir)

    def restore(
        self,
        snapshot_hash: str,
        restore_hashes: Dict[str, Optional[str]],
        restore_resources: List[ws.SnapshotResourceMixin],
    ) -> None:
        """We override restore to perform a git-fat pull at the end,
        if needed.
        """
        super().restore(snapshot_hash, restore_hashes, restore_resources)
        run_git_fat_pull_if_needed(self.workspace_dir, self.verbose)

    def remove_tag_from_snapshot(self, hash_val: str, tag: str) -> None:
        """Remove the specified tag from the specified snapshot. Throw an
        InternalError if either the snapshot or the tag do not exist.
        """
        md_filename = join(
            join(self.workspace_dir, SNAPSHOT_METADATA_DIR_PATH), "%s_md.json" % hash_val.lower()
        )
        if not exists(md_filename):
            raise InternalError("No metadata entry for snapshot %s" % hash_val)
        with open(md_filename, "r") as f:
            data = json.load(f)
        md = ws.SnapshotMetadata.from_json(data)
        assert md.hashval == hash_val
        if tag not in md.tags:
            raise InternalError("Tag %s not found in snapshot %s" % (tag, hash_val))
        md.tags = [tag for tag in md.tags if tag != tag]
        with open(md_filename, "w") as f:
            json.dump(md.to_json(), f, indent=2)

    def save_snapshot_metadata_and_manifest(
        self, metadata: SnapshotMetadata, manifest: bytes
    ) -> None:
        snapshot_dir_path = join(self.workspace_dir, SNAPSHOT_DIR_PATH)
        if not exists(snapshot_dir_path):
            os.makedirs(snapshot_dir_path)
        snapshot_manifest_path = join(snapshot_dir_path, "snapshot-%s.json" % metadata.hashval)
        with open(snapshot_manifest_path, "wb") as f:
            f.write(manifest)
        snapshot_md_dir = join(self.workspace_dir, SNAPSHOT_METADATA_DIR_PATH)
        if not exists(snapshot_md_dir):
            os.makedirs(snapshot_md_dir)
        snapshot_metadata_path = join(snapshot_md_dir, "%s_md.json" % metadata.hashval)
        with open(snapshot_metadata_path, "w") as mdf:
            json.dump(metadata.to_json(), mdf, indent=2)

    def as_snapshot_ws(self) -> ws.SnapshotWorkspaceMixin:
        """If this workspace supports snapshots, cast
        it to a SnapshotWorkspaceMixin. Otherwise,
        raise an NotSupportedError exception.
        """
        return cast(ws.SnapshotWorkspaceMixin, self)

    def as_lineage_ws(self) -> ws.SnapshotWorkspaceMixin:
        """If this workspace supports snapshots and lineage, cast
        it to a SnapshotWorkspaceMixin. Otherwise,
        raise an NotSupportedError exception.
        """
        return cast(ws.SnapshotWorkspaceMixin, self)

    def get_dws_git_branch(self) -> str:
        """This is an implementation-specific method to get the branch name for
        the workspace. It is used internally as well as by the Git subdirectory
        resource (which uses the same git repo as the workspace). We define this
        method rather than just getting the dws_git_branch parameter because it
        might not have been set if the repo was created pre-1.5."""
        branch = self.global_params.get(DWS_GIT_BRANCH, None)
        if branch is not None:
            return branch
        else:
            (dws_git_branch, _) = get_branch_info(self.workspace_dir, self.verbose)
            self.global_params[DWS_GIT_BRANCH] = dws_git_branch # cache for subsequent calls
            return dws_git_branch


class WorkspaceFactory(ws.WorkspaceFactory):
    @staticmethod
    def load_workspace(batch: bool, verbose: bool, parsed_uri: ParseResult) -> ws.Workspace:  # type: ignore
        path = parsed_uri.path
        if not isabs(path):
            path = abspath(expanduser(path))
        if not isdir(path):
            raise ConfigurationError("Workspace directory %s does not exist" % path)
        metadata_path = join(path, ".dataworkspace")
        if not isdir(metadata_path):
            raise ConfigurationError(
                "Workspace directory %s does not correspond to an initialized git-backend workspace"
                % path
            )
        return Workspace(path, batch, verbose)

    @staticmethod
    def init_workspace(  # type: ignore
        workspace_name: str,
        dws_version: str,
        global_params: JSONDict,
        local_params: JSONDict,
        batch: bool,
        verbose: bool,
        scratch_dir: str,
        workspace_dir: str,
        git_fat_remote: Optional[str] = None,
        git_fat_user: Optional[str] = None,
        git_fat_port: Optional[int] = None,
        git_fat_attributes: Optional[str] = None,
        git_lfs_attributes: Optional[str] = None,
    ) -> ws.Workspace:
        if not exists(workspace_dir):
            raise ConfigurationError(
                "Directory for new workspace '%s' does not exist" % workspace_dir
            )
        md_dir = join(workspace_dir, BASE_DIR)
        if isdir(md_dir):
            raise ConfigurationError(
                "Found %s directory under %s" % (BASE_DIR, workspace_dir)
                + " Has this workspace already been initialized?"
            )
        verify_git_config_initialized(workspace_dir, batch=batch, verbose=verbose)
        os.mkdir(md_dir)
        snapshot_dir = join(workspace_dir, SNAPSHOT_DIR_PATH)
        os.mkdir(snapshot_dir)
        snapshot_md_dir = join(workspace_dir, SNAPSHOT_METADATA_DIR_PATH)
        os.mkdir(snapshot_md_dir)

        (abs_scratch_dir, scratch_dir_gitignore) = init_scratch_directory(
            scratch_dir, workspace_dir, global_params, local_params
        )
        if exists(join(workspace_dir, ".git")):
            click.echo("%s is already a git repository, will just add to it" % workspace_dir)
        else:
            git_init(workspace_dir, verbose=verbose)
        with open(join(workspace_dir, RESOURCES_FILE_PATH), "w") as f:
            json.dump([], f, indent=2)
        with open(join(workspace_dir, LOCAL_PARAMS_PATH), "w") as f:
            json.dump(local_params, f, indent=2)
        with open(join(workspace_dir, RESOURCE_LOCAL_PARAMS_PATH), "w") as f:
            json.dump({}, f, indent=2)
        os.mkdir(join(workspace_dir, CURRENT_LINEAGE_DIR_PATH))

        with open(join(workspace_dir, GIT_IGNORE_FILE_PATH), "a") as f:
            f.write("%s\n" % basename(LOCAL_PARAMS_PATH))
            f.write("%s\n" % basename(RESOURCE_LOCAL_PARAMS_PATH))
            f.write("current_lineage/\n")
        git_add(
            workspace_dir,
            [RESOURCES_FILE_PATH, GIT_IGNORE_FILE_PATH],
            verbose=verbose,
        )
        if scratch_dir_gitignore is not None:
            # add the scratch directory's gitignore entry to the top level of
            # the repo, not the .gitignore within .dataworkspace
            ensure_entry_in_gitignore(
                workspace_dir, ".gitignore", scratch_dir_gitignore, commit=False, verbose=verbose
            )
        # common things that should not be committed
        ensure_entry_in_gitignore(
            workspace_dir, ".gitignore", ".ipynb_checkpoints/", commit=False, verbose=verbose
        )
        ensure_entry_in_gitignore(
            # Jupyter notebook autosave?
            workspace_dir,
            ".gitignore",
            ".~*",
            commit=False,
            verbose=verbose,
        )
        commit_changes_in_repo(workspace_dir, "dws init", verbose=verbose)

        # Now that we've done an initial commit, we figure out the branch we
        # are on, create the config file and commit that.
        (dws_git_branch, _) = get_branch_info(workspace_dir, verbose=verbose)
        global_params[DWS_GIT_BRANCH] = dws_git_branch
        with open(join(workspace_dir, CONFIG_FILE_PATH), "w") as f:
            json.dump(
                {
                    "name": workspace_name,
                    "dws-version": dws_version,
                    "global_params": global_params,
                },
                f,
                indent=2,
            )
        git_add(workspace_dir, [CONFIG_FILE_PATH], verbose=verbose)
        commit_changes_in_repo(workspace_dir, "dws init (config.json)", verbose=verbose)

        if not isdir(abs_scratch_dir):
            if verbose:
                print("Creating scratch directory %s" % abs_scratch_dir)
            os.makedirs(abs_scratch_dir)

        if git_fat_remote is not None:
            setup_git_fat_for_repo(
                workspace_dir,
                git_fat_remote,
                git_fat_user,
                git_fat_port,
                git_fat_attributes,
                verbose,
            )
        if git_lfs_attributes or is_a_git_lfs_repo(workspace_dir):
            if git_fat_remote:
                raise ConfigurationError("Cannot have both git-lfs and git-fat for the same repo.")
            init_git_lfs(workspace_dir, git_lfs_attributes, verbose=verbose)
        return Workspace(workspace_dir, batch, verbose)

    @staticmethod
    def clone_workspace(local_params: JSONDict, batch: bool, verbose: bool, *args) -> ws.Workspace:
        # args is REPOSITORY_URL [DIRECTORY]
        if len(args) == 0:
            raise ConfigurationError(
                "Need to specify a Git repository URL when cloning a workspace"
            )
        else:
            repository = args[0]  # type: str
        directory = args[1] if len(args) == 2 else None  # type: Optional[str]
        if len(args) > 2:
            raise ConfigurationError(
                "Clone of git backend expecting at most two arguments, received: %s" % repr(args)
            )

        # initial checks on the directory
        if directory:
            directory = abspath(expanduser(directory))
            parent_dir = dirname(directory)
            if isdir(directory):
                raise ConfigurationError("Clone target directory '%s' already exists" % directory)
            initial_path = directory
        else:
            parent_dir = abspath(expanduser(curdir))
            initial_path = join(
                parent_dir, uuid.uuid4().hex
            )  # get a unique name within this directory
        if not isdir(parent_dir):
            raise ConfigurationError("Parent directory '%s' does not exist" % parent_dir)
        if not os.access(parent_dir, os.W_OK):
            raise ConfigurationError("Unable to write into directory '%s'" % parent_dir)

        verify_git_config_initialized(parent_dir, batch=batch, verbose=verbose)

        # ping the remote repo
        cmd = [GIT_EXE_PATH, "ls-remote", "--quiet", repository]
        try:
            call_subprocess(cmd, parent_dir, verbose)
        except Exception as e:
            raise ConfigurationError("Unable to access remote repository '%s'" % repository) from e

        # we have to clone the repo first to find out its name!
        try:
            cmd = [GIT_EXE_PATH, "clone", repository, initial_path]
            call_subprocess(cmd, parent_dir, verbose)
            config_file = join(initial_path, CONFIG_FILE_PATH)
            if not exists(config_file):
                raise ConfigurationError("Did not find configuration file in remote repo")
            with open(config_file, "r") as f:
                config_json = json.load(f)
            if "name" not in config_json:
                raise InternalError("Missing 'name' property in configuration file")
            workspace_name = config_json["name"]
            if directory is None:
                new_name = join(parent_dir, workspace_name)
                if isdir(new_name):
                    raise ConfigurationError("Clone target directory %s already exists" % new_name)
                safe_rename(initial_path, new_name)
                directory = new_name

            cf_path = join(directory, CONFIG_FILE_PATH)
            if not exists(cf_path):
                raise ConfigurationError("Did not find workspace config file %s" % cf_path)
            with open(cf_path, "r") as f:
                cf_data = json.load(f)
            global_params = cf_data["global_params"]
            # get the scratch directory (also adds local param if needed)
            abs_scratch_dir = clone_scratch_directory(directory, global_params, local_params, batch)
            if DWS_GIT_BRANCH in global_params and (global_params[DWS_GIT_BRANCH] is not None):
                # if the branch is specified, make sure we are on it
                switch_git_branch_if_needed(directory, global_params[DWS_GIT_BRANCH],
                                            verbose=verbose)
            if not isdir(abs_scratch_dir):
                if verbose:
                    print("Creating scratch directory %s" % abs_scratch_dir)
                os.makedirs(abs_scratch_dir)
            with open(join(directory, LOCAL_PARAMS_PATH), "w") as f:
                json.dump(local_params, f, indent=2)  # create an initial local params file
            with open(join(directory, RESOURCE_LOCAL_PARAMS_PATH), "w") as f:
                json.dump(
                    {}, f, indent=2
                )  # create resource local params, to be populated via resource clones
            snapshot_md_dir = join(directory, SNAPSHOT_METADATA_DIR_PATH)
            if not exists(snapshot_md_dir):
                # It is possible that we are cloning a repo with no snapshots
                os.mkdir(snapshot_md_dir)
            snapshot_dir = join(directory, SNAPSHOT_DIR_PATH)
            if not exists(snapshot_dir):
                # It is possible that we are cloning a repo with no snapshots
                os.mkdir(snapshot_dir)
            current_lineage_dir = join(directory, CURRENT_LINEAGE_DIR_PATH)
            if not exists(current_lineage_dir):
                os.mkdir(current_lineage_dir)
            if is_a_git_fat_repo(directory):
                validate_git_fat_in_path()
                import dataworkspaces.third_party.git_fat as git_fat

                python2_exe = git_fat.find_python2_exe()
                git_fat.run_git_fat(python2_exe, ["init"], cwd=directory, verbose=verbose)
                # pull the objects referenced by the current head
                git_fat.run_git_fat(python2_exe, ["pull"], cwd=directory, verbose=verbose)
            ensure_git_lfs_configured_if_needed(directory, verbose=verbose)

        except:
            if isdir(initial_path):
                shutil.rmtree(initial_path)
            if (directory is not None) and isdir(directory):
                shutil.rmtree(directory)
            raise

        return WorkspaceFactory.load_workspace(batch, verbose, urlparse(directory))


FACTORY = WorkspaceFactory()
